# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
import re
from .common import cat
from .command import Command
# Fix connection refused for python 2.7
try:
    FileNotFoundError
except NameError:
    FileNotFoundError = IOError


MODULES = ['cuDNN', 'TensorRT.', 'VPI']  # 'Visionworks'
CUDA_FILE_RE = re.compile(r'CUDA Version (.*)')
CUDA_NVCC_RE = re.compile(r'V([0-9]+.[0-9]+.[0-9]+)')


def get_cuda():
    cuda_version = ''
    if os.path.isfile("/usr/local/cuda/version.txt"):
        version = cat("/usr/local/cuda/version.txt")
        match = re.search(CUDA_FILE_RE, version)
        if match:
            cuda_version = match.group(1)
    elif os.path.isfile("/usr/local/cuda/bin/nvcc"):
        cmd = Command(['/usr/local/cuda/bin/nvcc', '--version'])
        try:
            lines = cmd()
            for line in lines:
                match = re.search(CUDA_NVCC_RE, line)
                if match:
                    cuda_version = match.group(1)
                    break
        except (OSError, Command.CommandException):
            pass
    return cuda_version


def get_opencv():
    opencv_version = ''
    opencv_cuda = False
    cmd = Command(['opencv_version'])
    try:
        lines = cmd()
        # Extract OpenCV version
        opencv_version = ''.join(lines)
        # Extract if compiled with CUDA
        cmd = Command(['opencv_version', '--verbose'])
        lines = cmd()
        for line in lines:
            if "NVIDIA CUDA" in line:
                opencv_cuda = True
                break
            if "Use Cuda" in line:
                opencv_cuda = False if "NO" in line else True
                break
    except (OSError, Command.CommandException):
        pass
    return opencv_version, opencv_cuda


def get_all_modules():
    modules = {}
    # Extract all modules in dpkg -l
    dpkg = Command(['dpkg', '-l'])
    try:
        lines = dpkg()
        for row in lines:
            row = re.sub(r'\n+ +', '\n', row)  # remove spaces at the start of lines and empty lines
            row = re.sub(r'\s +', '\t', row)  # replace two or more spaces with tab
            cells = row.split('\t')
            if len(cells) > 2:
                name = cells[1]
                version = cells[2]
                if version not in ['arm64', 'amd64']:
                    modules[name] = version
    except (OSError, Command.CommandException):
        pass
    return modules


def get_libraries():
    os_variables = {}
    # Find all modules
    modules = get_all_modules()
    for name in MODULES:
        # Fix TensorRT search #462
        name_dict = name[:-1] if name.endswith('.') else name
        os_variables[name_dict] = ''
        # Find version if installed
        for module, version in modules.items():
            if name.endswith('.') and name.lower()[:-1] == module:
                os_variables[name_dict] = version.split('-')[0]
                break
            if name.lower() in module:
                os_variables[name_dict] = version.split('-')[0]
                break
    # Get Vulkan output
    cmd_vulkaninfo = Command(['which', 'vulkaninfo'])
    try:
        lines = cmd_vulkaninfo()
        # Extract version
        if not lines:
            raise Command.CommandException("Missing command", -3)
        cmd_vulkan = Command(lines)
        lines = cmd_vulkan()
        for line in lines:
            if "Vulkan Instance Version" in line:
                os_variables['Vulkan'] = line.lstrip("Vulkan Instance Version: ")
                break
    except (OSError, Command.CommandException):
        pass
    return os_variables


if __name__ == "__main__":
    # Get CUDA
    cuda_version = get_cuda()
    if cuda_version:
        print("CUDA: {version}".format(version=cuda_version))
    else:
        print("CUDA not installed!")
    # Find OpenCV
    opencv_version, opencv_cuda = get_opencv()
    if opencv_version:
        opencv_cuda_string = "YES" if opencv_cuda else "NO"
        print("OPENCV: {opencv_version} - with CUDA: {opencv_cuda}".format(opencv_version=opencv_version, opencv_cuda=opencv_cuda_string))
    else:
        print("OPENCV not Installed!")
    # Get all libraries
    os_variables = get_libraries()
    for name, value in os_variables.items():
        print("{name}: {value}".format(name=name, value=value))
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import re
import os
import pwd
from .common import cat
# Logging
import logging
# Create logger
logger = logging.getLogger(__name__)

MEM_TABLE_REG = re.compile(r'^(?P<user>\w+)\s+(?P<process>[^ ]+)\s+(?P<PID>\d+)\s+(?P<size>\d+)(?P<unit>\w)\n')
TOT_TABLE_REG = re.compile(r'total\s+(?P<size>\d+)(?P<unit>\w)')


def read_process_table(path_table):
    """
    This method list all processes working with GPU

    ========== ============ ======== =============
    user       process      PID      size
    ========== ============ ======== =============
    user       name process number   dictionary
    ========== ============ ======== =============

    :return: list of all processes
    :type spin: list
    """
    table = []
    total = {}
    with open(path_table, "r") as fp:
        for line in fp:
            # Search line
            match = re.search(MEM_TABLE_REG, line)
            if match:
                parsed_line = match.groupdict()
                data = [
                    parsed_line['PID'],
                    parsed_line['user'],
                    parsed_line['process'],
                    int(parsed_line['size']),
                ]
                table += [data]
                continue
            # Find total on table
            match = re.search(TOT_TABLE_REG, line)
            if match:
                parsed_line = match.groupdict()
                total = int(parsed_line['size'])
                continue
    # return total and table
    return total, table


class ProcessService(object):

    def __init__(self):
        self.usernames = {4294967295: "root"}
        # board type
        self._root_path = "/sys/kernel"
        if os.getenv('JTOP_TESTING', False):
            self._root_path = "/fake_sys/kernel"
            logger.warning("Running in JTOP_TESTING folder={root_dir}".format(root_dir=self._root_path))
        # Check if jetson board
        self._isJetson = os.path.isfile(self._root_path + "/debug/nvmap/iovmm/maps")
        # Get the clock ticks per second and page size
        self._clk_tck = os.sysconf('SC_CLK_TCK')
        # self._page_size = os.sysconf('SC_PAGE_SIZE')
        # Initialization memory
        logger.info("Process service started")

    def get_process_info(self, pid, gpu_mem_usage, process_name, uptime):
        # Check if exist folder
        if not os.path.isdir(os.path.join('/proc', pid)):
            return []
        # https://man7.org/linux/man-pages/man5/proc.5.html
        stat = cat(os.path.join('/proc', pid, 'stat')).split()
        # Decode uid and find username
        uid = int(cat(os.path.join('/proc', pid, 'loginuid')))
        if uid not in self.usernames:
            self.usernames[uid] = pwd.getpwuid(uid).pw_name
        # Read memory process
        # Extract resident set size (VmRSS) (Second field)
        # VmRSS is the resident set size of the process, which is the portion of the process's memory
        # that is held in RAM and is not swapped out to disk. This is the amount of memory that the process is currently using.
        mem_raw = cat(os.path.join('/proc', pid, 'statm')).split()
        vm_rss = int(mem_raw[1])
        # CPU percent
        # https://stackoverflow.com/questions/16726779/how-do-i-get-the-total-cpu-usage-of-an-application-from-proc-pid-stat
        utime = float(stat[13])
        stime = float(stat[14])
        starttime = float(stat[21]) / self._clk_tck
        total_time = (utime + stime) / self._clk_tck
        proc_uptime = uptime - starttime
        cpu_percent = 100 * (total_time / proc_uptime)

        process = [
            int(pid),               # pid process
            self.usernames[uid],    # username
            "I",                    # GPU name
            "Graphic",              # type process
            int(stat[17]),          # Priority
            stat[2],                # state
            cpu_percent,            # CPU percent
            vm_rss,                 # MEM process
            gpu_mem_usage,          # GPU mem usage
            process_name,           # Process name
        ]
        return process

    def get_status(self):
        total = {}
        table = []
        # Update table
        if self._isJetson:
            # Use the memory table to measure
            total, table = read_process_table(self._root_path + "/debug/nvmap/iovmm/maps")

            uptime = float(open('/proc/uptime', 'r').readline().split()[0])

            table = [self.get_process_info(prc[0], prc[3], prc[2], uptime) for prc in table]

        return total, table
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import curses
from .jtopgui import Page
# Graphics elements
from .lib.common import NColors, plot_name_info, size_min, unit_to_string, size_to_string
from .lib.chart import Chart
from .lib.process_table import ProcessTable
from .lib.linear_gauge import basic_gauge, freq_gauge
from .lib.smallbutton import SmallButton
from .pcontrol import color_temperature


def gpu_gauge(stdscr, pos_y, pos_x, size, gpu_data, idx):
    gpu_status = gpu_data['status']
    # Data gauge
    data = {
        'name': 'GPU' if idx == 0 else 'GPU{idx}'.format(idx=idx),
        'color': NColors.green() | curses.A_BOLD,
        'values': [(gpu_status['load'], NColors.igreen())],
    }
    if 'freq' in gpu_data:
        # Draw current frequency
        curr_string = unit_to_string(gpu_data['freq']['cur'], 'k', 'Hz')
        stdscr.addstr(pos_y, pos_x + size - 8, curr_string, NColors.italic())
    # Draw gauge
    basic_gauge(stdscr, pos_y, pos_x, size - 10, data, bar=" ")


def compact_gpu(stdscr, pos_y, pos_x, width, jetson):
    line_counter = 0
    # Status all GPUs
    if jetson.gpu:
        for idx, gpu in enumerate(jetson.gpu.values()):
            gpu_gauge(stdscr, pos_y + line_counter, pos_x, width, gpu, idx)
            line_counter += 1
    else:
        data = {
            'name': 'GPU',
            'color': NColors.green() | curses.A_BOLD,
            'online': False,
            'coffline': NColors.igreen(),
            'message': 'NVIDIA GPU NOT DETECTED/AVAILABLE',
        }
        basic_gauge(stdscr, pos_y, pos_x, width - 2, data)
        line_counter = 1
    return line_counter


class GPU(Page):

    def __init__(self, stdscr, jetson):
        super(GPU, self).__init__("GPU", stdscr, jetson)
        # Check if grey exist otherwise use white
        COLOR_GREY = 240 if curses.COLORS >= 256 else curses.COLOR_WHITE
        # Initialize GPU chart
        self.draw_gpus = {}
        for gpu_name in self.jetson.gpu:
            type_gpu = "i" if self.jetson.gpu[gpu_name]['type'] == 'integrated' else 'd'
            chart = Chart(jetson, "{t}GPU {name}".format(t=type_gpu, name=gpu_name), self.update_chart,
                          color_text=curses.COLOR_GREEN)
            # button_railgate = SmallButton(stdscr, self.action_railgate, info={'name': gpu_name})
            button_3d_scaling = SmallButton(stdscr, self.action_scaling_3D, info={'name': gpu_name})
            if type_gpu == 'i':
                chart_ram = Chart(jetson, "GPU Shared RAM", self.update_chart_ram,
                                  type_value=float,
                                  color_text=curses.COLOR_GREEN,
                                  color_chart=[COLOR_GREY, curses.COLOR_GREEN])
            else:
                chart_ram = None
            self.draw_gpus[gpu_name] = {'chart': chart, '3d_scaling': button_3d_scaling, 'ram': chart_ram}
        # Add Process table
        self.process_table = ProcessTable(self.stdscr, self.jetson)

    def action_railgate(self, info, selected):
        # Read current status railgate
        status_railgate = not self.jetson.gpu.get_railgate(info['name'])
        # Set new railgate
        self.jetson.gpu.set_railgate(info['name'], status_railgate)

    def action_scaling_3D(self, info, selected):
        # Read current status 3d_scaling
        status_3d_scaling = not self.jetson.gpu.get_scaling_3D(info['name'])
        # Set new 3d_scaling
        self.jetson.gpu.set_scaling_3D(info['name'], status_3d_scaling)

    def update_chart(self, jetson, name):
        # Decode GPU name
        gpu_name = name.split(" ")[1]
        gpu_data = jetson.gpu[gpu_name]
        gpu_status = gpu_data['status']
        # Append in list
        return {
            'value': [gpu_status['load']],
        }

    def update_chart_ram(self, jetson, name):
        parameter = jetson.memory['RAM']
        # Get max value if is present
        max_val = parameter.get("tot", 100)
        # Get value
        cpu_val = parameter.get("used", 0)
        use_val = parameter.get("shared", 0)
        szw, divider, unit = size_min(max_val, start='k')
        # Append in list
        used_out = (cpu_val) / divider
        gpu_out = (use_val) / divider
        return {
            'value': [used_out, gpu_out],
            'max': szw,
            'unit': unit
        }

    def draw(self, key, mouse):
        # Screen size
        height, width, first = self.size_page()
        # Measure height
        gpu_height = (height * 2 // 3 - 3) // len(self.jetson.gpu)
        # Plot all GPU temperatures
        self.stdscr.addstr(first + 1, 1, "Temperatures:", curses.A_NORMAL)
        for idx, name in enumerate(self.jetson.temperature):
            if 'gpu' in name.lower():
                sensor = self.jetson.temperature[name]
                color_temperature(self.stdscr, first + 1, 15, name, sensor)
        # Draw all GPU
        for idx, (gpu_name, gpu_data) in enumerate(self.jetson.gpu.items()):
            chart = self.draw_gpus[gpu_name]['chart']
            chart_ram = self.draw_gpus[gpu_name]['ram']
            gpu_status = gpu_data['status']
            gpu_freq = gpu_data['freq']
            # Set size chart gpu
            size_x = [1, width // 2 - 2]
            size_y = [first + 2 + idx * (gpu_height + 1), first + 2 + (idx + 1) * (gpu_height - 3)]
            # Print status CPU
            governor = gpu_freq.get('governor', '')
            label_chart_gpu = "{percent: >3.0f}% - gov: {governor}".format(percent=gpu_status['load'], governor=governor)
            # Draw GPU chart
            chart.draw(self.stdscr, size_x, size_y, label=label_chart_gpu)
            # Draw GPU RAM chart
            size_x_ram = [1 + width // 2, width - 2]
            mem_data = self.jetson.memory['RAM']
            total = size_to_string(mem_data['tot'], 'k')
            shared = size_to_string(mem_data['shared'], 'k')
            chart_ram.draw(self.stdscr, size_x_ram, size_y, label="{used}/{total}B".format(used=shared, total=total))
            # Print all status GPU
            button_position = width // 4
            button_idx = 0
            # 3D scaling
            scaling_string = "Active" if gpu_status['3d_scaling'] else "Disable"
            scaling_status = NColors.green() if gpu_status['3d_scaling'] else curses.A_NORMAL
            try:
                self.stdscr.addstr(first + 1 + (idx + 1) * gpu_height - 1, 1 + button_idx, "3D scaling:", curses.A_BOLD)
                self.draw_gpus[gpu_name]['3d_scaling'].update(first + 1 + (idx + 1) * gpu_height - 1, 12 + button_idx,
                                                              scaling_string, key=key, mouse=mouse, color=scaling_status)
            except curses.error:
                pass
            button_idx += button_position
            # railgate status
            railgate_string = "Active" if gpu_status['railgate'] else "Disable"
            railgate_status = NColors.green() if gpu_status['railgate'] else curses.A_NORMAL
            plot_name_info(self.stdscr, first + 1 + (idx + 1) * gpu_height - 1, 1 + button_idx, "Railgate", railgate_string, color=railgate_status)
            # self.stdscr.addstr(first + 1 + (idx + 1) * gpu_height - 1, 1 + button_idx, "Railgate:", curses.A_BOLD)
            # self.draw_gpus[gpu_name]['railgate'].update(first + 1 + (idx + 1) * gpu_height - 1, 10 + button_idx, railgate_string,
            #                                             key=key, mouse=mouse, color=railgate_status)
            button_idx += button_position
            # Power control
            plot_name_info(self.stdscr, first + 1 + (idx + 1) * gpu_height - 1, 1 + button_idx, "Power ctrl", gpu_data['power_control'])
            button_idx += button_position
            # TPC PG Mask
            if 'tpc_pg_mask' in gpu_status:
                tpc_pg_mask_string = "ON" if gpu_status['tpc_pg_mask'] else "OFF"
                # tpc_pg_mask_status = NColors.green() if gpu_status['tpc_pg_mask'] else NColors.red()
                plot_name_info(self.stdscr, first + 1 + (idx + 1) * gpu_height - 1, 1 + button_idx, "TPC PG", tpc_pg_mask_string)
                button_idx += button_position
            # Check if GPC data is included
            frq_size = width - 3
            if 'GPC' in gpu_freq:
                size_gpc_gauge = (width - 2) // (2 + len(gpu_freq['GPC']))
                for gpc_idx, gpc in enumerate(gpu_freq['GPC']):
                    freq_data = {
                        'name': 'GPC{idx}'.format(idx=gpc_idx),
                        'cur': gpc,
                        'unit': 'k',
                        'online': gpc > 0,
                    }
                    freq_gauge(self.stdscr, first + 1 + (idx + 1) * gpu_height, width // 2 + gpc_idx * (size_gpc_gauge) + 2, size_gpc_gauge - 1, freq_data)
                # Change size frequency GPU
                frq_size = width // 2
            # Print frequency info
            gpu_freq['name'] = "Frq"
            freq_gauge(self.stdscr, first + 1 + (idx + 1) * gpu_height, 1, frq_size, gpu_freq)
        # Draw all Processes
        height_table = height - first + 2 + gpu_height
        self.process_table.draw(first + 2 + gpu_height, 0, width, height_table, key, mouse)
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
import re
import argparse
from .core.common import get_var
from .core.nvpmodel import nvpmodel_query
from .core.hardware import get_platform_variables
from .core.jetson_variables import get_jetson_variables
from .core.jetson_libraries import get_libraries, get_cuda, get_opencv
from .core.exceptions import JtopException
from .service import status_service
from .terminal_colors import bcolors
# Version match
VERSION_RE = re.compile(r""".*__version__ = ["'](.*?)['"]""", re.S)
COPYRIGHT_RE = re.compile(r""".*__copyright__ = ["'](.*?)['"]""", re.S)


def main():
    parser = argparse.ArgumentParser(
        description='Show detailed information about this board. Machine, Jetpack, libraries and other',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-v', '--verbose', dest="verbose", help='Show all variables', action="store_true", default=False)
    parser.add_argument('-s', '--serial', dest="serial", help='Show serial number', action="store_true", default=False)
    # Copyrights
    print("Software part of jetson-stats {version} - {copyright}".format(version=get_var(VERSION_RE), copyright=get_var(COPYRIGHT_RE)))
    # Parse arguments
    args = parser.parse_args()
    # Read all Jetson Variables
    jetson = get_jetson_variables()
    # Extract and remove Serial Number
    serial_number = jetson['Serial Number']
    del jetson['Serial Number']
    # Print headline
    if jetson['Jetpack']:
        print("Model: {model} - Jetpack {jetpack} [L4T {L4T}]".format(model=bcolors.bold(jetson['Model']),
              jetpack=bcolors.bold(jetson['Jetpack']), L4T=bcolors.bold(jetson['L4T'])))
    else:
        print(bcolors.fail(bcolors.bold("Jetpack missing!")))
        print(" - Model: {model}".format(model=bcolors.bold(jetson['Model'])))
        print(" - L4T: {L4T}".format(L4T=bcolors.bold(jetson['L4T'])))
    del jetson['Model']
    del jetson['Jetpack']
    del jetson['L4T']
    # Print NVP model status
    try:
        nvpmodel_now = nvpmodel_query()
        print("{service}[{number}]: {name}".format(
            service=bcolors.ok(bcolors.bold("NV Power Mode")),
            name=bcolors.bold(nvpmodel_now['name']),
            number=bcolors.bold(nvpmodel_now['id'])))
    except JtopException:
        pass
    # Print serial number
    if serial_number:
        if not args.serial:
            serial_number = "[XXX Show with: jetson_release -s XXX]"
        print("{sn_string} {serial_number}".format(sn_string=bcolors.ok(bcolors.bold("Serial Number:")), serial_number=serial_number))
    # Print jetson hardware variables
    if not args.verbose:
        for name in ['699-level Part Number', 'SoC', 'Codename', 'BoardIDs', 'CUDA Arch BIN']:
            if name in jetson:
                del jetson[name]
    print(bcolors.ok(bcolors.bold("Hardware:")))
    for name, variable in jetson.items():
        if not variable:
            variable = bcolors.fail("Not available")
            if name == 'Module' and os.getuid() != 0:
                variable = bcolors.fail("Check with sudo")
        print(" - {name}: {variable}".format(name=bcolors.bold(name), variable=variable))
    # Print platform variables
    plat = get_platform_variables()
    if not args.verbose:
        for name in ['Machine', 'System', 'Python']:
            if name in plat:
                del plat[name]
    print(bcolors.ok(bcolors.bold("Platform:")))
    for name, variable in plat.items():
        if not variable:
            variable = bcolors.fail("Not available")
        print(" - {name}: {variable}".format(name=bcolors.bold(name), variable=variable))
    # jtop status
    print(bcolors.ok(bcolors.bold("jtop:")))
    # Print status jetson-stats service
    print(" - {name}: {value}".format(name=bcolors.bold("Version"), value=get_var(VERSION_RE)))
    jtop_status = bcolors.ok("Active") if status_service() else bcolors.fail("Inactive")
    print(" - {name}: {value}".format(name=bcolors.bold("Service"), value=jtop_status))
    # Read CUDA status
    print(bcolors.ok(bcolors.bold("Libraries:")))
    cuda_version = get_cuda()
    if cuda_version:
        print(" - {name}: {value}".format(name=bcolors.bold('CUDA'), value=cuda_version))
    else:
        print(" - {name}: {value}".format(name=bcolors.bold('CUDA'), value=bcolors.fail("Not installed")))
    # Read all libraries
    os_variables = get_libraries()
    for name, value in os_variables.items():
        if not value:
            value = bcolors.fail("Not installed")
        print(" - {name}: {value}".format(name=bcolors.bold(name), value=value))
    # Read OpenCV status
    opencv_version, opencv_cuda = get_opencv()
    if opencv_version:
        opencv_cuda_string = bcolors.ok("YES") if opencv_cuda else bcolors.fail("NO")
        opencv_string = "{name}: {value}".format(name=bcolors.bold('OpenCV'), value=opencv_version)
        print(" - {opencv_string} - with CUDA: {opencv_cuda}".format(opencv_string=opencv_string, opencv_cuda=opencv_cuda_string))
    else:
        print(" - {name}: {value}".format(name=bcolors.bold('OpenCV'), value=bcolors.fail("Not installed")))


if __name__ == "__main__":
    main()
# EOF



#!/usr/bin/python
# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
import re
import sys
import warnings
import logging
from shutil import copyfile
try:
    from smbus2 import SMBus  # pyright: ignore[reportMissingImports]
except ImportError:
    print("Skip for setup.py")

from .common import cat
from .command import Command
# Create logger
logger = logging.getLogger(__name__)
# https://github.com/robotframework/robotframework/issues/2552
if not sys.warnoptions:
    warnings.filterwarnings('ignore', category=RuntimeWarning, module='runpy')

# ---------------------
# JETPACK DETECTION
# ---------------------
# Write version of jetpack installed
# https://developer.nvidia.com/embedded/jetpack-archive
NVIDIA_JETPACK = {
    # -------- JP6 --------
    "36.2.0": "6.0 DP",
    # -------- JP5 --------
    "35.4.1": "5.1.2",
    "35.3.1": "5.1.1",
    "35.3.0": "5.1.1 PRE",
    "35.2.1": "5.1",
    "35.1.0": "5.0.2 GA",
    "34.1.1": "5.0.1 DP",
    "34.1.0": "5.0 DP",
    "34.0.1": "5.0 PRE-DP",
    # -------- JP4 --------
    "32.7.4": "4.6.4",
    "32.7.3": "4.6.3",
    "32.7.2": "4.6.2",
    "32.7.1": "4.6.1",
    "32.6.1": "4.6",
    "32.5.2": "4.5.1",
    "32.5.1": "4.5.1",
    "32.5.0": "4.5",
    "32.5": "4.5",
    "32.4.4": "4.4.1",
    "32.4.3": "4.4",
    "32.4.2": "4.4 DP",
    "32.3.1": "4.3",
    "32.2.3": "4.2.3",
    "32.2.1": "4.2.2",
    "32.2.0": "4.2.1",
    "32.2": "4.2.1",
    "32.1.0": "4.2",
    "32.1": "4.2",
    "31.1.0": "4.1.1",
    "31.1": "4.1.1",
    "31.0.2": "4.1",
    "31.0.1": "4.0",
    # -------- Old JP --------
    "28.4.0": "3.3.3",
    "28.2.1": "3.3 | 3.2.1",
    "28.2.0": "3.2",
    "28.2": "3.2",
    "28.1.0": "3.1",
    "28.1": "3.1",
    "27.1.0": "3.0",
    "27.1": "3.0",
    "24.2.1": "3.0 | 2.3.1",
    "24.2.0": "2.3",
    "24.2": "2.3",
    "24.1.0": "2.2.1 | 2.2",
    "24.1": "2.2.1 | 2.2",
    "23.2.0": "2.1",
    "23.2": "2.1",
    "23.1.0": "2.0",
    "23.1": "2.0",
    "21.5.0": "2.3.1 | 2.3",
    "21.5": "2.3.1 | 2.3",
    "21.4.0": "2.2 | 2.1 | 2.0 | 1.2 DP",
    "21.4": "2.2 | 2.1 | 2.0 | 1.2 DP",
    "21.3.0": "1.1 DP",
    "21.3": "1.1 DP",
    "21.2.0": "1.0 DP",
    "21.2": "1.0 DP",
}


CUDA_TABLE = {
    'tegra234': '8.7',  # JETSON ORIN - tegra234
    'tegra23x': '8.7',  # JETSON ORIN - tegra234
    'tegra194': '7.2',  # JETSON XAVIER
    'tegra186': '6.2',  # JETSON TX2
    'tegra210': '5.3',  # JETSON TX1
    'tegra124': '3.2',  # JETSON TK1
}

# Module reference
# https://docs.nvidia.com/jetson/archives/l4t-archived/l4t-3231/index.html
# https://docs.nvidia.com/jetson/archives/r35.2.1/DeveloperGuide/text/IN/QuickStart.html
MODULE_NAME_TABLE = {
    'p3767-0005': 'NVIDIA Jetson Orin Nano (Developer kit)',
    'p3767-0004': 'NVIDIA Jetson Orin Nano (4GB ram)',
    'p3767-0003': 'NVIDIA Jetson Orin Nano (8GB ram)',
    'p3767-0001': 'NVIDIA Jetson Orin NX (8GB ram)',
    'p3767-0000': 'NVIDIA Jetson Orin NX (16GB ram)',
    'p3701-0005': 'NVIDIA Jetson AGX Orin (64GB ram)',
    'p3701-0004': 'NVIDIA Jetson AGX Orin (32GB ram)',
    'p3701-0002': 'NVIDIA Jetson IGX Orin (Developer kit)',
    'p3701-0000': 'NVIDIA Jetson AGX Orin',
    'p3668-0003': 'NVIDIA Jetson Xavier NX (16GB ram)',
    'p3668-0001': 'NVIDIA Jetson Xavier NX',
    'p3668-0000': 'NVIDIA Jetson Xavier NX (Developer kit)',
    'p2888-0008': 'NVIDIA Jetson AGX Xavier Industrial (32 GB ram)',
    'p2888-0006': 'NVIDIA Jetson AGX Xavier (8 GB ram)',
    'p2888-0005': 'NVIDIA Jetson AGX Xavier (64 GB ram)',
    'p2888-0004': 'NVIDIA Jetson AGX Xavier (32 GB ram)',
    'p2888-0003': 'NVIDIA Jetson AGX Xavier (32 GB ram)',
    'p2888-0001': 'NVIDIA Jetson AGX Xavier (16 GB ram)',
    'p3448-0003': 'NVIDIA Jetson Nano (2 GB ram)',
    'p3448-0002': 'NVIDIA Jetson Nano module (16Gb eMMC)',
    'p3448-0000': 'NVIDIA Jetson Nano (4 GB ram)',
    'p3636-0001': 'NVIDIA Jetson TX2 NX',
    'p3509-0000': 'NVIDIA Jetson TX2 NX',
    'p3489-0888': 'NVIDIA Jetson TX2 (4 GB ram)',
    'p3489-0000': 'NVIDIA Jetson TX2i',
    'p3310-1000': 'NVIDIA Jetson TX2',
    'p2180-1000': 'NVIDIA Jetson TX1',
    'r375-0001': 'NVIDIA Jetson TK1',
    'p3904-0000': 'NVIDIA Clara AGX',
    # Other modules
    'p2595-0000-A0': 'Nintendo Switch'
}
# ---------------------
# DO NOT EDIT FROM HERE
# ---------------------
# THIS SCRIPT MUST HAVE .SH !
JTOP_VARIABLE_FILE = 'jtop_env.sh'
DTSFILENAME_RE = re.compile(r'(.*)-p')
SOC_RE = re.compile(r'[0-9]+')
DPKG_L4T_CORE_RE = re.compile(r'^nvidia-l4t-core.*install$')
# Number 7 is for Jetson TX2
I2C_EEPROM_BUS = [0, 1, 2, 7]
RAW_FILES = ['/etc/nv_tegra_release',
             '/sys/firmware/devicetree/base/model',
             '/proc/device-tree/nvidia,boardids',
             '/proc/device-tree/compatible',
             '/proc/device-tree/nvidia,dtsfilename']


def read_i2c_raw_data(bus, registry, size_block):
    string_data = ""
    for idx in range(256 // size_block):
        # Extract part number expected module
        data = bus.read_i2c_block_data(registry, idx * size_block, size_block)
        string_data += ' '.join(["{:02X}".format(x) for x in data]) + " " * 4
        string_data += ''.join([chr(x) if chr(x).isprintable() else "." for x in data]) + "\n"
    return string_data


def get_jetson_raw_output():
    raw_output = {}
    # Catch all output from all files
    for file in RAW_FILES:
        raw_output[file] = cat(file).strip('\n') if os.path.isfile(file) else "No such file or directory"
    # Read all output from all I2C ports
    for bus_number in I2C_EEPROM_BUS:
        try:
            # Forcing read SMBus added from Jetpack 6.0 DP
            bus = SMBus(bus_number, force=True)
            size_block = 16
            raw_output['I2C-{num}-0x50'.format(num=bus_number)] = read_i2c_raw_data(bus, 0x50, size_block)
            raw_output['I2C-{num}-0x56'.format(num=bus_number)] = read_i2c_raw_data(bus, 0x56, size_block)
        except (IOError, OSError):
            # print("Error I2C bus: {bus_number}".format(bus_number=bus_number))
            raw_output['I2C-{num}'.format(num=bus_number)] = 'FAIL'
    return raw_output


def check_dpkg_nvidia_l4t_core():
    dpkg = Command(['dpkg', '--get-selections'])
    lines = dpkg()
    for line in lines:
        if re.match(DPKG_L4T_CORE_RE, line):
            return True
    return False


def get_nvidia_l4t():
    # Read NV TEGRA RELEASE
    if os.path.isfile('/etc/nv_tegra_release'):
        # NVIDIA Jetson version
        # reference https://devtalk.nvidia.com/default/topic/860092/jetson-tk1/how-do-i-know-what-version-of-l4t-my-jetson-tk1-is-running-/
        # https://stackoverflow.com/questions/16817646/extract-version-number-from-a-string
        # https://askubuntu.com/questions/319307/reliably-check-if-a-package-is-installed-or-not
        # https://github.com/dusty-nv/jetson-inference/blob/7e81381a96c1ac5f57f1728afbfdec7f1bfeffc2/tools/install-pytorch.sh#L296
        nv_tegra_release = cat("/etc/nv_tegra_release").split(", ")
        l4t_release = nv_tegra_release[0].lstrip("# R").rstrip(" (release)")
        l4t_revision = nv_tegra_release[1].lstrip("REVISION: ")
        return '.'.join([l4t_release, l4t_revision])
        # Ectract GCID - DO NOT NEEDED
        # os_variables['GCID'] = nv_tegra_release[2].lstrip("GCID: ")
        # Ectract SOC - DO NOT USE THIS LINE! CONTAINS ALWAYS WRONG OUTPUT
        # number = re.search(SOC_RE, nv_tegra_release[3].lstrip("BOARD: ")).group()
        # os_variables['SOC'] = "tegra{number}".format(number=number)
    elif check_dpkg_nvidia_l4t_core():
        dpkg = Command(['dpkg-query', '--showformat=\'${Version}\'', '--show', 'nvidia-l4t-core'])
        l4t = dpkg()[0]
        return l4t.split('-')[0].lstrip('\'')
    # If not find any L4T return empty string
    return ''


def get_variables_from_dtsfilename():
    hardware = {}
    # Decode dtsfilename
    if os.path.isfile("/proc/device-tree/nvidia,dtsfilename"):
        # Read dtsfilename
        # AGX Orin - tegra234-p3701-0000-p3737-0000
        # Nano - tegra210-p3448-0000-p3449-0000-b00
        # TX2 - tegra186-quill-p3310-1000-c03-00-base
        # TX1 - tegra210-jetson-tx1-p2597-2180-a01-devkit
        # TK1 - tegra124-jetson_tk1-pm375-000-c00-00
        dtsfilename = cat("/proc/device-tree/nvidia,dtsfilename").split('/')
        # Decode codename
        hardware['CODENAME'] = dtsfilename[-3]
        # Decode NVIDIA Jetson type, model and board
        jetson_soc_module_board = dtsfilename[-1].rstrip('.dts').split('-')
        hardware['SOC'] = jetson_soc_module_board[0]
        # print("type: {jetson_type}".format(jetson_type=hardware['TYPE']))
        parts = '-'.join(jetson_soc_module_board[1:])
        match = re.match(DTSFILENAME_RE, parts)
        if match:
            module = match.group(1)
            hardware['PART_NUMBER'] = module
            # print(f"module: {module}".format(module=module))
            carrier = parts.replace("{module}-".format(module=module), '')
            hardware['CARRIER'] = carrier
            # print(f"carrier: {carrier}".format(carrier=carrier))
            # Decode Jetson type of module
            # https://docs.nvidia.com/jetson/archives/r34.1/DeveloperGuide/index.html
            hardware['MODULE'] = MODULE_NAME_TABLE.get(module, '')
        else:
            print("jetson model and board not available")
            print(parts)
    # Decode CUDA architecure
    hardware['CUDA_ARCH_BIN'] = CUDA_TABLE.get(hardware['SOC'], '')
    return hardware


def get_part_number():
    part_number = ''
    jetson_part_number = ''
    # Find 699-level part number from EEPROM and extract P-number
    for bus_number in I2C_EEPROM_BUS:
        try:
            # Forcing read SMBus added from Jetpack 6.0 DP
            bus = SMBus(bus_number, force=True)
            part_number = bus.read_i2c_block_data(0x50, 20, 29)
            part_number = ''.join(chr(i) for i in part_number).rstrip('\x00')
            # print(part_number)
            board_id = part_number[5:9]
            sku = part_number[10:14]
            jetson_part_number = "p{board_id}-{sku}".format(board_id=board_id, sku=sku)
            return part_number, jetson_part_number
        except (IOError, OSError):
            # logger.error("I2C[{bus_number}] [Errno {errno}]: {msg}".format(errno=e.errno, bus_number=bus_number, msg=e.strerror))
            pass
    return part_number, jetson_part_number


def get_jetson_variables():
    hardware = {}
    # Read Jetson model
    model = ''
    if os.path.isfile('/sys/firmware/devicetree/base/model'):
        model = cat("/sys/firmware/devicetree/base/model")
    hardware['Model'] = model
    # Find part number from I2C
    # https://docs.nvidia.com/jetson/archives/l4t-archived/l4t-3243/index.html
    # https://docs.nvidia.com/jetson/archives/r34.1/DeveloperGuide/text/HR/JetsonEepromLayout.html
    part_number, jetson_part_number = get_part_number()
    # Add only if there is a part number
    if part_number:
        hardware['699-level Part Number'] = part_number
        hardware['P-Number'] = jetson_part_number
    # Read boardids
    if os.path.isfile('/proc/device-tree/nvidia,boardids'):
        boardids = cat("/proc/device-tree/nvidia,boardids")
        boardids = "p{boardids}".format(boardids=boardids.replace(':', '-'))
        hardware['BoardIDs'] = boardids
    # Find module from part_number
    module = ''
    if jetson_part_number:
        module = MODULE_NAME_TABLE.get(jetson_part_number, '')
    elif 'BoardIDs' in hardware:
        module = MODULE_NAME_TABLE.get(hardware['BoardIDs'], '')
    hardware['Module'] = module
    # Decode SOC
    compatible = ''
    if os.path.isfile("/proc/device-tree/compatible"):
        compatible = cat("/proc/device-tree/compatible").split(',')
        compatible = compatible[-1]
    hardware['SoC'] = compatible
    # Decode CUDA architecture
    cuda_arch = ''
    for cuda, arch in CUDA_TABLE.items():
        if cuda in hardware['SoC']:
            cuda_arch = arch
            break
    hardware['CUDA Arch BIN'] = cuda_arch
    # Find codename
    if os.path.isfile("/proc/device-tree/nvidia,dtsfilename"):
        # Read dtsfilename
        # AGX Orin - tegra234-p3701-0000-p3737-0000
        # Nano - tegra210-p3448-0000-p3449-0000-b00
        # TX2 - tegra186-quill-p3310-1000-c03-00-base
        # TX1 - tegra210-jetson-tx1-p2597-2180-a01-devkit
        # TK1 - tegra124-jetson_tk1-pm375-000-c00-00
        dtsfilename = cat("/proc/device-tree/nvidia,dtsfilename")
        if 'kernel-dts' in dtsfilename:
            # Decode codename
            idx_codename = dtsfilename.split('/').index('kernel-dts') - 1
            hardware['Codename'] = dtsfilename.split('/')[idx_codename].capitalize()
        elif "/dvs/git/dirty/git-master_linux/kernel" in dtsfilename:
            # Decode codename for TK1
            hardware['Codename'] = dtsfilename.split('/')[-3].capitalize()
    # Read serial number
    serial = ''
    if os.path.isfile('/sys/firmware/devicetree/base/serial-number'):
        serial = cat("/sys/firmware/devicetree/base/serial-number")
    hardware['Serial Number'] = serial
    # Extract L4T
    hardware['L4T'] = get_nvidia_l4t()
    # Read Jetpack
    hardware['Jetpack'] = NVIDIA_JETPACK.get(hardware['L4T'], '')
    return hardware


def status_variables(name=JTOP_VARIABLE_FILE):
    return os.path.isfile('/etc/profile.d/{name}'.format(name=name))


def uninstall_variables(name=JTOP_VARIABLE_FILE):
    if os.path.isfile('/etc/profile.d/{name}'.format(name=name)):
        logger.info("Found {name}".format(name=name))
        # Remove old jetson_stats variable
        os.remove('/etc/profile.d/{name}'.format(name=name))
        logger.info(" - Remove {name} from /etc/profile.d/".format(name=name))


def install_variables(package_root, copy, name=JTOP_VARIABLE_FILE):
    logger.info("Install {name} variables".format(name=name))
    variables_install_path = '/etc/profile.d/{name}'.format(name=name)
    variables_package_path = '{package_root}/scripts/{name}'.format(package_root=package_root, name=name)
    # remove if exist file
    if os.path.isfile(variables_install_path) or os.path.islink(variables_install_path):
        logger.info(" - Remove old {path}".format(path=variables_install_path))
        os.remove(variables_install_path)
    if copy:
        type_service = "Copying"
        copyfile(variables_package_path, variables_install_path)
    else:
        type_service = "Linking"
        os.symlink(variables_package_path, variables_install_path)
    # Prompt message
    logger.info(" - {type} {file} -> {path}".format(type=type_service.upper(), file=name, path=variables_install_path))


def export_variables(hardware):
    # Export variables to be loaded on bash script
    # https://blog.tintoy.io/2017/06/exporting-environment-variables-from-python-to-bash/
    for name, value in hardware.items():
        # Escape symbols commonly used by Bash.
        value = value.replace('"', '\\"').replace('$', '\\$').replace('`', '\\`')
        name = name.upper().replace("-", "_").replace(" ", "_")
        print('export JETSON_{}="{}"'.format(name, value))


# If you want to run this script:
# sudo python3 -m jtop.core.jetson_variables
if __name__ == "__main__":
    # Get all variables
    hardware = get_jetson_variables()
    # Remove variables not needed
    if '699-level Part Number' in hardware:
        del hardware['699-level Part Number']
    # Test output variables
    export_variables(hardware)
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import curses


def init_colorscale_pair(num, fg, bg):
    curses.init_pair(num, fg if curses.COLORS >= 256 else curses.COLOR_WHITE, bg if curses.COLORS >= 256 else curses.COLOR_BLACK)


class NColors:

    RED = 1
    GREEN = 2
    YELLOW = 3
    BLUE = 4
    MAGENTA = 5
    CYAN = 6

    iRED = 7
    iGREEN = 8
    iYELLOW = 9
    iBLUE = 10
    iMAGENTA = 11
    iCYAN = 12

    def __init__(self, color_filter):
        # Define pairing colors
        curses.init_pair(NColors.RED, curses.COLOR_RED if not color_filter else curses.COLOR_BLUE, curses.COLOR_BLACK)
        curses.init_pair(NColors.GREEN, curses.COLOR_GREEN, curses.COLOR_BLACK)
        curses.init_pair(NColors.YELLOW, curses.COLOR_YELLOW, curses.COLOR_BLACK)
        curses.init_pair(NColors.BLUE, curses.COLOR_BLUE, curses.COLOR_BLACK)
        curses.init_pair(NColors.MAGENTA, curses.COLOR_MAGENTA, curses.COLOR_BLACK)
        curses.init_pair(NColors.CYAN, curses.COLOR_CYAN, curses.COLOR_BLACK)
        # background
        curses.init_pair(NColors.iRED, curses.COLOR_WHITE, curses.COLOR_RED if not color_filter else curses.COLOR_BLUE)
        curses.init_pair(NColors.iGREEN, curses.COLOR_WHITE, curses.COLOR_GREEN)
        curses.init_pair(NColors.iYELLOW, curses.COLOR_BLACK, curses.COLOR_YELLOW)
        curses.init_pair(NColors.iBLUE, curses.COLOR_WHITE, curses.COLOR_BLUE)
        curses.init_pair(NColors.iMAGENTA, curses.COLOR_WHITE, curses.COLOR_MAGENTA)
        curses.init_pair(NColors.iCYAN, curses.COLOR_WHITE, curses.COLOR_CYAN)

    @staticmethod
    def init_grey(num):
        init_colorscale_pair(num, 240, curses.COLOR_BLACK)

    @staticmethod
    def italic():
        # Check if Italic is included
        return curses.A_ITALIC if hasattr(curses, 'A_ITALIC') else curses.A_NORMAL

    @staticmethod
    def red():
        return curses.color_pair(NColors.RED)

    @staticmethod
    def green():
        return curses.color_pair(NColors.GREEN)

    @staticmethod
    def yellow():
        return curses.color_pair(NColors.YELLOW)

    @staticmethod
    def blue():
        return curses.color_pair(NColors.BLUE)

    @staticmethod
    def magenta():
        return curses.color_pair(NColors.MAGENTA)

    @staticmethod
    def cyan():
        return curses.color_pair(NColors.CYAN)

    @staticmethod
    def ired():
        return curses.color_pair(NColors.iRED)

    @staticmethod
    def igreen():
        return curses.color_pair(NColors.iGREEN)

    @staticmethod
    def iyellow():
        return curses.color_pair(NColors.iYELLOW)

    @staticmethod
    def iblue():
        return curses.color_pair(NColors.iBLUE)

    @staticmethod
    def imagenta():
        return curses.color_pair(NColors.iMAGENTA)

    @staticmethod
    def icyan():
        return curses.color_pair(NColors.iCYAN)
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
import re
import stat
import shlex
# Logging
import logging
import subprocess as sp
from .processes import read_process_table
from .engine import read_engine
from .common import cat, GenericInterface
from .command import Command
# Create logger
logger = logging.getLogger(__name__)
# Memory regular exception
MEMINFO_REG = re.compile(r'(?P<key>.+):\s+(?P<value>.+) (?P<unit>.)B')
FSTAB_RE = re.compile(r'^(?P<path>[^ ]+) +(?P<mount>[^ ]+) +(?P<type>[^ ]+) +(?P<options>[^ ]+) +(?P<dump>\d+) +(?P<pass>\d+)$')
BUDDYINFO_REG = re.compile(r'Node\s+(?P<numa_node>\d+).*zone\s+(?P<zone>\w+)\s+(?P<nr_free>.*)')
SWAP_REG = re.compile(r'(?P<name>[^ ]+)\s+(?P<type>[^ ]+)\s+(?P<size>\d+)\s+(?P<used>\d+)\s+(?P<prio>-?\d+)')
# Swap configuration
PATH_FSTAB = '/etc/fstab'
CONFIG_DEFAULT_SWAP_DIRECTORY = ''
CONFIG_DEFAULT_SWAP_NAME = 'swfile'


def meminfo():
    # Read meminfo and decode
    # https://access.redhat.com/solutions/406773
    status_mem = {}
    with open("/proc/meminfo", 'r') as fp:
        for line in fp:
            # Search line
            match = re.search(MEMINFO_REG, line.strip())
            if match:
                parsed_line = match.groupdict()
                status_mem[parsed_line['key']] = int(parsed_line['value'])
    return status_mem


def buddyinfo(page_size):
    # Read status free memory
    # http://andorian.blogspot.com/2014/03/making-sense-of-procbuddyinfo.html
    buddyhash = {}
    with open("/proc/buddyinfo", 'r') as fp:
        buddyinfo = fp.readlines()
    for line in buddyinfo:
        # Decode line
        parsed_line = re.match(BUDDYINFO_REG, line.strip()).groupdict()
        # detect buddy size
        numa_node = int(parsed_line["numa_node"])
        free_fragments = [int(i) for i in parsed_line["nr_free"].split()]
        max_order = len(free_fragments)
        fragment_sizes = [page_size * 2**order for order in range(0, max_order)]
        usage_in_bytes = [free * fragmented for free, fragmented in zip(free_fragments, fragment_sizes)]
        data = {
            "zone": parsed_line["zone"],
            "nr_free": free_fragments,
            "sz_fragment": fragment_sizes,
            "usage": usage_in_bytes}
        buddyhash[numa_node] = buddyhash[numa_node] + [data] if numa_node in buddyhash else [data]
    return buddyhash


def read_swapon():
    """
    This method list all processes working with GPU

    ============== ======================= ======== =============
    name           type                    prio     size
    ============== ======================= ======== =============
    name partition type: partition or file priority dictionary
    ============== ======================= ======== =============

    :return: list of all processes
    :type spin: list
    """
    table = {}
    swap = Command(['swapon', '--show', '--raw', '--byte'])
    lines = swap()
    for line in lines:
        # Search line
        match = re.search(SWAP_REG, line.strip())
        if match:
            parsed_line = match.groupdict()
            name = parsed_line['name']
            data = {
                # Improve this detection. (Now only checking if zram is on name to saving time)
                'type': parsed_line['type'] if 'zram' not in name else 'zram',
                'prio': int(parsed_line['prio']),
                'size': int(parsed_line['size']) // 1024,
                'used': int(parsed_line['used']) // 1024,
            }
            table[name] = data
    return table


def read_fstab():
    fstab = {}
    with open(PATH_FSTAB, "r") as fp:
        for line in fp:
            match = re.search(FSTAB_RE, line.strip())
            if match:
                parsed_line = match.groupdict()
                path = parsed_line['path']
                del parsed_line['path']
                fstab[path] = parsed_line
    return fstab


def read_emc(root_path):
    emc = {}
    if os.path.isdir(root_path + "/debug/bpmp/debug/clk/emc"):
        path = root_path + "/debug/bpmp/debug/clk/emc"
        # Check if access to this file
        if os.access(path + "/rate", os.R_OK):
            with open(path + "/rate", 'r') as f:
                # Write min
                emc['cur'] = int(f.read()) // 1000
        # Check if access to this file
        if os.access(path + "/max_rate", os.R_OK):
            with open(path + "/max_rate", 'r') as f:
                # Write min
                emc['max'] = int(f.read()) // 1000
        # Check if access to this file
        if os.access(path + "/min_rate", os.R_OK):
            with open(path + "/min_rate", 'r') as f:
                # Write min
                emc['min'] = int(f.read()) // 1000
        # Check if access to this file
        if os.access(path + "/mrq_rate_locked", os.R_OK):
            with open(path + "/mrq_rate_locked", 'r') as f:
                # Write min
                emc['override'] = int(f.read()) // 1000
    elif os.path.isdir(root_path + "/debug/tegra_bwmgr"):
        path = root_path + "/debug/clk/override.emc"
        # Check if access to this file
        if os.access(path + "/clk_rate", os.R_OK):
            with open(path + "/clk_rate", 'r') as f:
                # Write min
                emc['cur'] = int(f.read()) // 1000
        # Check if access to this file
        if os.access(path + "/clk_state", os.R_OK):
            with open(path + "/clk_state", 'r') as f:
                # Write min
                emc['override'] = int(f.read()) // 1000
        # Decode from tegra_bwmgr
        path = root_path + "/tegra_bwmgr"
        # Check if access to this file
        if os.access(path + "/emc_max_rate", os.R_OK):
            with open(path + "/emc_max_rate", 'r') as f:
                # Write min
                emc['max'] = int(f.read()) // 1000
        # Check if access to this file
        if os.access(path + "/emc_min_rate", os.R_OK):
            with open(path + "/emc_min_rate", 'r') as f:
                # Write min
                emc['min'] = int(f.read()) // 1000
    elif os.path.isdir(root_path + "/clk/emc"):
        emc = read_engine(root_path + "/clk/emc")
    # Fix max frequency
    emc_cap = 0
    # Check if access to this file
    if os.access(root_path + "/nvpmodel_emc_cap/emc_iso_cap", os.R_OK):
        with open(root_path + "/nvpmodel_emc_cap/emc_iso_cap", 'r') as f:
            # Write min
            emc_cap = int(f.read()) // 1000
    # Fix max EMC
    if 'max' in emc:
        if emc_cap > 0 and emc_cap < emc['max']:
            emc['max'] = emc_cap
    # Percentage utilization
    # https://forums.developer.nvidia.com/t/real-time-emc-bandwidth-with-sysfs/107479/3
    if os.access(root_path + "/debug/cactmon/mc_all", os.R_OK):
        with open(root_path + "/debug/cactmon/mc_all", 'r') as f:
            utilization = int(f.read())
    elif os.access(root_path + "/actmon_avg_activity/mc_all", os.R_OK):
        with open(root_path + "/actmon_avg_activity/mc_all", 'r') as f:
            utilization = int(f.read())
    else:
        # if utilization not accesibile return empty EMC data
        return {}
    emc['val'] = utilization // emc['cur']
    # Set always online this engine
    emc['online'] = True
    return emc


class Memory(GenericInterface):
    """
    This class get the output from your memory, this class is readable like a dictionary,
    please read the documentation on :py:attr:`~jtop.jtop.memory` but is also usable to enable, disable swap on your device

    .. code-block:: python

        with jtop() as jetson:
            if jetson.ok():
                jetson.memory.swap_set(10, on_boot=False)


    or if you want to deactivate a swap you can use this command

    .. code-block:: python

        with jtop() as jetson:
            if jetson.ok():
                jetson.memory.swap_deactivate()

    Below all methods available using the :py:attr:`~jtop.jtop.memory` attribute
    """

    def __init__(self):
        super(Memory, self).__init__()

    def swap_path(self):
        """
        Return the default SWAP path, Default path `/`

        :return: Path swap
        :rtype: str
        """
        return self._init

    def clear_cache(self):
        """
        Clear the memory cache
        """
        # Set new swap size configuration
        self._controller.put({'clear_cache': ''})

    def swap_is_enable(self, path):
        """
        Check if a swap is on list

        :param path: Path swap
        :type path: str

        :return: Status swap
        :rtype: bool
        """
        return path in self._data['SWAP']['table']

    def swap_set(self, value, path='', on_boot=False):
        """
        Create a new swap on a default path. :py:func:`~swap_path`

        :param value: Size in **G** of a new SWAP
        :type value: int
        :param path: Path swap, defaults ''
        :type path: str, optional
        :param on_boot: Set this swap on boot, defaults False
        :type on_boot: bool, optional
        :raises ValueError: value is not an :py:class:`int` or a :py:class:`float`
        """
        if not isinstance(value, (int, float)):
            raise ValueError("Need a Number")
        # if path_swap is empty load from default configuration
        if not path:
            path = self._init
        # Set new swap size configuration
        self._controller.put({'swap': {'command': 'set', 'path': path, 'size': value, 'boot': on_boot}})

    def swap_deactivate(self, path=''):
        """
        Deactivate a swap from a path or from default location `/`

        :param path: Path swap
        :type path: str, optional
        """
        # if path_swap is empty load from default configuration
        if not path:
            path = self._init
        # Set new swap size configuration
        self._controller.put({'swap': {'command': 'unset', 'path': path}})


class MemoryService(object):

    def __init__(self, config):
        self._config = config
        # Extract memory page size
        self._page_size = os.sysconf("SC_PAGE_SIZE")
        # board type
        self._root_path = "/sys/kernel"
        if os.getenv('JTOP_TESTING', False):
            self._root_path = "/fake_sys/kernel"
            logger.warning("Running in JTOP_TESTING folder={root_dir}".format(root_dir=self._root_path))
        self._is_emc = True if read_emc(self._root_path) else False
        if self._is_emc:
            logger.info("Found EMC!")
        self._is_iram = os.path.isdir(self._root_path + "/debug/nvmap/iram")
        if self._is_iram:
            logger.info("Found IRAM!")
        # Initialization memory
        logger.info("Memory service started")

    def swap_path(self):
        config = self._config.get('swap', {})
        directory = config.get('directory', CONFIG_DEFAULT_SWAP_DIRECTORY)
        swap_name = config.get('name', CONFIG_DEFAULT_SWAP_NAME)
        return "{directory}/{name}".format(directory=directory, name=swap_name)

    def clear_cache(self):
        """
        Clear cache following https://coderwall.com/p/ef1gcw/managing-ram-and-swap
        """
        clear_cache = Command(['sysctl', 'vm.drop_caches=3'])
        out = clear_cache()
        return True if out else False

    @staticmethod
    def swap_set(size, path_swap, on_boot):
        if os.path.isfile(path_swap):
            logger.error("{path_swap} already exist".format(path_swap=path_swap))
            return
        # Load swap configuration
        logger.info("Activate {path_swap} auto={on_boot}".format(path_swap=path_swap, on_boot=on_boot))
        # Create a swapfile for Ubuntu at the current directory location
        sp.call(shlex.split('fallocate -l {size}G {path_swap}'.format(size=size, path_swap=path_swap)))
        # Change permissions so that only root can use it
        # https://www.tutorialspoint.com/python/os_chmod.htm
        # Equivalent permission 600 srw-------
        os.chmod(path_swap, stat.S_IREAD | stat.S_IWRITE)
        # Set up the Linux swap area
        sp.call(shlex.split('mkswap {path_swap}'.format(path_swap=path_swap)))
        # Now start using the swapfile
        sp.call(shlex.split('swapon {path_swap}'.format(path_swap=path_swap)))
        # Add not on boot return
        if not on_boot:
            return
        # Find if is already on boot
        swap_string_boot = "{path_swap} none swap sw 0 0".format(path_swap=path_swap)
        fstab = read_fstab()
        if path_swap in fstab:
            logger.warn("{path_swap} Already on boot".format(path_swap=path_swap))
            return
        # Append swap line
        file_object = open(PATH_FSTAB, 'a')
        file_object.write("{swap_string_boot}\n".format(swap_string_boot=swap_string_boot))
        file_object.close()

    @staticmethod
    def swap_deactivate(path_swap):
        # Check if exist swap
        if not os.path.isfile(path_swap):
            logger.error("{path_swap} Does not exist".format(path_swap=path_swap))
            return
        # Disable swap
        sp.call(shlex.split('swapoff {path_swap}'.format(path_swap=path_swap)))
        # Remove swap
        os.remove(path_swap)
        # Run script
        logger.info("Deactivate {path_swap}".format(path_swap=path_swap))
        # Remove if on fstab
        swap_string_boot = "{path_swap} none swap sw 0 0".format(path_swap=path_swap)
        fstab = read_fstab()
        if path_swap not in fstab:
            return
        # Check if is on boot
        logger.info("Removing {path_swap} from fstab".format(path_swap=path_swap))
        with open(PATH_FSTAB, "r") as f:
            lines = f.readlines()
        with open(PATH_FSTAB, "w") as f:
            for line in lines:
                if line.strip("\n") != swap_string_boot:
                    f.write(line)
        # Run script
        logger.info("Removed {path_swap} from boot".format(path_swap=path_swap))

    def get_status(self, mem_total):
        memory = {}
        # Measure the largest free bank for 4MB
        mem_size = buddyinfo(self._page_size)
        # Count only the biggest Large free bank (lfb)
        large_free_bank = 0
        for _, data in mem_size.items():
            large_free_bank += sum([zone['nr_free'][-1] for zone in data])
        # Status Memory
        status_mem = meminfo()
        # Read memory use
        # NvMapMemUsed: Is the shared memory between CPU and GPU
        # This key is always available on Jetson (not really always)
        ram_shared = status_mem.get('NvMapMemUsed', 0)
        if mem_total:
            # Update shared size
            ram_shared = mem_total if ram_shared == 0 else ram_shared
        # Extract memory info
        ram_total = status_mem.get('MemTotal', 0)
        ram_free = status_mem.get('MemFree', 0)
        # ram_available = status_mem.get('MemAvailable', 0)
        ram_buffer = status_mem.get('Buffers', 0)
        ram_cached = status_mem.get('Cached', 0)
        ram_SReclaimable = status_mem.get('SReclaimable', 0)
        # ram_Shmem = status_mem.get('Shmem', 0)
        total_used_memory = ram_total - ram_free
        cached_memory = ram_cached + ram_SReclaimable  # + ram_Shmem
        # Add fields for RAM
        memory['RAM'] = {
            'tot': ram_total,
            'used': total_used_memory - (ram_buffer + ram_cached),
            'free': ram_free,
            'buffers': ram_buffer,
            'cached': cached_memory,
            'shared': ram_shared,
            'lfb': large_free_bank,  # In 4MB
        }
        # Extract swap numbers
        swap_total = status_mem.get('SwapTotal', 0)
        swap_free = status_mem.get('SwapFree', 0)
        swap_cached = status_mem.get('SwapCached', 0)
        # read all swap and check if are on boot
        swap_table = read_swapon()
        fstab = read_fstab()
        for name in swap_table:
            swap_table[name]['boot'] = name in fstab
        # Add fields for swap
        memory['SWAP'] = {
            'tot': swap_total,
            'used': swap_total - swap_free,
            'cached': swap_cached,
            'table': swap_table,
        }
        # Read EMC status
        if self._is_emc:
            memory['EMC'] = read_emc(self._root_path)
        # Read IRAM if available
        if self._is_iram:
            size = 0
            if os.path.isfile(self._root_path + "/debug/nvmap/iram/size"):
                # Convert from Hex to decimal - Number in bytes
                size = int(cat(self._root_path + "/debug/nvmap/iram/size"), 16) // 1024
            used_total, _ = read_process_table(self._root_path + "/debug/nvmap/iram/clients")
            memory['IRAM'] = {
                'tot': size,
                'used': used_total,
                'lfb': size - used_total,  # TODO To check
            }
        return memory
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import re
import os
import sys
import argparse
from .terminal_colors import bcolors
from .core.memory import MemoryService, read_swapon
from .core.common import get_var
# Version match
VERSION_RE = re.compile(r""".*__version__ = ["'](.*?)['"]""", re.S)
COPYRIGHT_RE = re.compile(r""".*__copyright__ = ["'](.*?)['"]""", re.S)


def main():
    parser = argparse.ArgumentParser(
        description='Create a swap file and enable on boot (require sudo)',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-d', '--dir', dest="directory", help='Directory to place swapfile', type=str, default='')
    parser.add_argument('-n', '--name', dest="name", help='Name swap file', type=str, default='swapfile')
    parser.add_argument('-s', '--size', dest="size", help='Size in Gigabytes', type=int, default='8')
    parser.add_argument('-a', '--auto', dest="auto", help='Enable swap on boot', action="store_true", default=False)
    parser.add_argument('-t', '--status', dest="status", help='Check if the swap is currently active', action="store_true", default=False)
    parser.add_argument('--off', dest="off", help='Switch off the swap', action="store_true", default=False)
    # Parse arguments
    args = parser.parse_args()
    # Copyrights
    print("Software part of jetson-stats {version} - {copyright}".format(version=get_var(VERSION_RE), copyright=get_var(COPYRIGHT_RE)))
    # Status swap
    if args.status:
        # Print all swap
        swap_table = read_swapon()
        print("NAME\t\tTYPE\t\tPRIO\t\tSIZE\t\tUSED")
        for name, swap in swap_table.items():
            print("{name}\t{type}\t\t{prio}\t\t{size}{unit}\t{used}{unit}".format(
                name=name,
                type=swap['type'],
                prio=swap['prio'],
                size=swap['size'],
                used=swap['used'],
                unit='k'))
        sys.exit(0)
    # Check if running a root
    if os.getuid() != 0:
        # Quit with error
        print(bcolors.fail("Please run with sudo"))
        parser.print_help(sys.stderr)
        sys.exit(1)
    # Define Memory Service
    memory_service = MemoryService
    # Path swap
    size = args.size
    auto = args.auto
    path_swap = "{directory}/{name}".format(directory=args.directory, name=args.name)
    if args.off:
        print("Switch off swap {path_swap}".format(path_swap=path_swap))
        memory_service.swap_deactivate(path_swap)
        sys.exit(0)
    # Create Swap
    print("Create swap {path_swap} [{size}GB] - on boot={auto}".format(path_swap=path_swap, size=size, auto=auto))
    # Create swap
    memory_service.swap_set(size, path_swap, auto)


if __name__ == "__main__":
    main()
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import re
import signal
import os
import sys
# import locale
import argparse
# control command line
import curses
# Logging
import logging
# jtop service
from .service import JtopServer
# jtop client
from .jtop import jtop
# jtop exception
from .core.exceptions import JtopException
from .core.common import get_var
# GUI jtop interface
from .jetson_config import jtop_config
from .gui import JTOPGUI, ALL, GPU, CPU, ENGINE, MEM, CTRL, INFO, engine_model
# Load colors
from .terminal_colors import bcolors
from .github import jetpack_missing, hardware_missing, engine_gui, get_hardware_log
# Create logger
logger = logging.getLogger(__name__)
# Version match
VERSION_RE = re.compile(r""".*__version__ = ["'](.*?)['"]""", re.S)
# Reference repository
REPOSITORY = "https://github.com/rbonghi/jetson_stats/issues"
LOOP_SECONDS = 5
JTOP_LOG_NAME = 'jtop-error.log'


def warning_messages(jetson, no_warnings=False):
    if no_warnings:
        return
    # Read status version
    hardware = jetson.board['hardware']
    version = get_var(VERSION_RE)
    # Check is well stored the default jetson_clocks configuration
    if jetson.jetson_clocks:
        if not jetson.jetson_clocks.is_config:
            print("[{status}] Please stop manually jetson_clocks or reboot this board".format(status=bcolors.warning()))
    # Check if an hardware value is missing
    if not all([data for name, data in hardware.items() if name not in ['Jetpack']]):
        hardware_missing(REPOSITORY, hardware, version)
    # Check if jetpack is missing
    if not hardware['Jetpack'] and hardware['L4T']:
        jetpack_missing(REPOSITORY, hardware, version)
    # Check if model is in map list
    if not engine_model(hardware["Module"]) and hardware["Module"]:
        engine_gui(REPOSITORY, hardware, version)


def exit_signal(signum, frame):
    logger.info("Close service by signal {signum}".format(signum=signum))
    sys.exit(0)


def main():
    parser = argparse.ArgumentParser(
        description='jtop is system monitoring utility and runs on terminal',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--force', dest='force', help=argparse.SUPPRESS, action="store_true", default=False)
    parser.add_argument('--health', dest="health", help='Status jtop and fix', action="store_true", default=False)
    parser.add_argument('--error-log', dest="log", help='Generate a log for GitHub', action="store_true", default=False)
    parser.add_argument('--no-warnings', dest="no_warnings", help='Do not show warnings', action="store_true", default=False)
    parser.add_argument('--restore', dest="restore", help='Reset Jetson configuration', action="store_true", default=False)
    parser.add_argument('--loop', dest="loop", help='Automatically switch page every {sec}s'.format(sec=LOOP_SECONDS), action="store_true", default=False)
    parser.add_argument('--color-filter', dest="color_filter",
                        help='Change jtop base colors, you can use also JTOP_COLOR_FILTER=True', action="store_true", default=False)
    parser.add_argument('-r', '--refresh', dest="refresh", help='refresh interval', type=int, default='1000')
    parser.add_argument('-p', '--page', dest="page", help='Open fix page', type=int, default=1)
    parser.add_argument('-v', '--version', action='version', version='%(prog)s {version}'.format(version=get_var(VERSION_RE)))
    # Parse arguments
    args = parser.parse_args()
    # Initialize signals
    # signal.signal(signal.SIGINT, exit_signal)  # Do not needed equivalent to exception KeyboardInterrupt
    signal.signal(signal.SIGTERM, exit_signal)
    # Run jtop service
    if os.getenv('JTOP_SERVICE', False):
        # Initialize logging level
        logging.basicConfig(level=logging.INFO, filemode='w', format='[%(levelname)s] %(name)s - %(message)s')
        # Run service
        try:
            # Initialize stats server
            server = JtopServer(force=args.force)
            server.loop_for_ever()
        except JtopException as e:
            print(e)
        # Close service
        sys.exit(0)
    # Initialize logging level
    logging.basicConfig()
    # Convert refresh to second
    interval = float(args.refresh / 1000.0)
    # Restore option
    if args.restore:
        try:
            with jtop(interval=interval) as jetson:
                # Write warnings
                if 'L4T' in jetson.board['hardware']:
                    warning_messages(jetson, args.no_warnings)
                # Restore configuration
                if jetson.ok():
                    for status, name in jetson.restore():
                        status = bcolors.ok() if status else bcolors.fail()
                        print(" [{status}] {name}".format(name=name, status=status))
        except JtopException as e:
            print(e)
        # Close service
        sys.exit(0)
    # Run health jtop
    if args.health:
        jtop_config()
    # Generate a log for GitHub
    if args.log:
        body = get_hardware_log()
        body += "\n\nLog from jtop {version}\n".format(version=get_var(VERSION_RE))
        with open('{cwd}/{name}'.format(cwd=os.getcwd(), name=JTOP_LOG_NAME), 'w') as writer:
            writer.write(body)
        print("LOG '{name}' generated in {path}".format(name=JTOP_LOG_NAME, path=os.getcwd()))
        sys.exit(0)
    # jtop client start
    try:
        # https://stackoverflow.com/questions/56373360/n-curses-within-python-how-to-catch-and-print-non-ascii-character
        # Commented for issues #466 #393
        # locale.setlocale(locale.LC_ALL, '')
        # Open jtop client
        with jtop(interval=interval) as jetson:
            # Call the curses wrapper
            color_filter = bool(os.getenv('JTOP_COLOR_FILTER', args.color_filter))
            # Build list pages available
            pages = [ALL]
            if jetson.gpu:
                pages += [GPU]
            pages += [CPU, MEM]
            if jetson.engine:
                pages += [ENGINE]
            if jetson.fan or jetson.jetson_clocks is not None or jetson.nvpmodel is not None:
                pages += [CTRL]
            pages += [INFO]
            curses.wrapper(JTOPGUI, jetson, pages, init_page=args.page,
                           loop=args.loop, seconds=LOOP_SECONDS, color_filter=color_filter)
            # Write warnings
            if 'L4T' in jetson.board['hardware']:
                warning_messages(jetson, args.no_warnings)
    except (KeyboardInterrupt, SystemExit):
        pass
    except JtopException as e:
        print(e)


if __name__ == "__main__":
    main()
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

# Math functions
import itertools
from math import ceil
import curses
from collections import deque
from .common import check_curses


class Chart(object):

    COLOR_COUNTER = 0
    OFFSET_COLOR_CHART = 20
    OFFSET_COLOR_TEXT = 18

    """
    Chart draw object
    http://www.melvilletheatre.com/articles/ncurses-extended-characters/index.html
    """

    def __init__(self, jetson, name, callback, type_value=int, line="*", color_text=curses.COLOR_WHITE, color_chart=[], fill=True, time=10.0, tik=2):
        self._color_obj_counter = Chart.COLOR_COUNTER
        self.jetson = jetson
        self.name = name
        self.callback = callback
        # Set shape
        self.refresh = jetson.interval * 1000
        # Design chart shape
        self.line = line
        self.color_text = color_text
        self.color_chart = color_chart if color_chart else [color_text]
        self.fill = fill
        # Set timing
        self.time = time
        self.tik = tik
        # Initialization chart
        max_record = int(self.time * (float(1.0 / float(self.refresh)) * 1000.0))
        self.values = deque(max_record * [(len(self.color_chart) * [0])], maxlen=max_record)
        # Initialize default values and unit
        self.unit = "%"
        self.type_value = type_value
        self.max_val = 100
        self.active = True
        self.message = "OFF"
        # Initialize all colors
        color_step = len(self.color_chart)
        values = list(range(len(self.color_chart)))[::-1] + [len(self.color_chart)]
        self._combinations = list(itertools.combinations(values, 2))
        list_colors = {}
        for c in self._combinations:
            if c[0] not in list_colors:
                list_colors[c[0]] = []
            list_colors[c[0]] = [c] + list_colors[c[0]]
        try:
            for idx, list_element in enumerate(list_colors.values()):
                list_element = sorted(list_element, key=lambda x: x[1], reverse=True)
                for idy, color_set in enumerate(list_element):
                    idx_name = Chart.OFFSET_COLOR_CHART + self._color_obj_counter + (len(self.color_chart) - idx - 1) * color_step + idy
                    second_color = self.color_chart[color_set[1]] if color_set[1] < len(self.color_chart) else curses.COLOR_BLACK
                    curses.init_pair(idx_name, self.color_chart[color_set[0]], second_color)
        except curses.error:
            curses.use_default_colors()
        # Update counter colors
        Chart.COLOR_COUNTER += len(self._combinations) + 1
        # Attach the chart for every update from jtop
        jetson.attach(self.update)

    def __del__(self):
        # Remove from color counter
        Chart.COLOR_COUNTER -= len(self._combinations) + 1

    def statusChart(self, active, message):
        self.active = active
        self.message = message

    def update(self, jetson):
        """ Local update chart """
        # Call the callback
        data = self.callback(jetson, self.name)
        # Get max value if is present
        self.max_val = data.get("max", self.max_val)
        # Get unit
        self.unit = data.get("unit", self.unit)
        # Get status
        self.active = data.get("active", self.active)
        # update the queue
        value = data.get("value", [0])
        self.values.append(value)

    def draw_y_axis(self, stdscr, pos_y, pos_x, size_height):
        self._plot_y_axis(stdscr, [0, pos_x + 4], [pos_y, pos_y + size_height])

    @check_curses
    def draw(self, stdscr, size_x, size_y, label="", y_label=True):
        curses.init_pair(Chart.OFFSET_COLOR_TEXT, self.color_text, curses.COLOR_BLACK)
        # Evaluate Diplay X, and Y size
        displayX = size_x[1] - size_x[0] + 1
        # displayY = size_y[1] - size_y[0] - 1
        # Text label
        stdscr.addstr(size_y[0], size_x[0], self.name, curses.A_BOLD)
        if label:
            stdscr.addstr(size_y[0], size_x[0] + len(self.name) + 1, label[:displayX - len(self.name)],
                          curses.color_pair(Chart.OFFSET_COLOR_TEXT) | curses.A_BOLD)
        # Draw ticks and labels
        self._plot_x_axis(stdscr, size_x, size_y, label=y_label)
        # Plot chart shape and labels
        self._plot_y_axis(stdscr, size_x, size_y, label=y_label)
        # Plot chart lines
        # Plot values
        self._plot_values(stdscr, size_x, size_y, label=y_label)
        # Add messsage not active
        if not self.active:
            l_label = size_x[1] - 6 if y_label else size_x[1] - 1
            # rectangle(stdscr, size_y[0] + 1, size_x[0], size_y[1], l_label)
            stdscr.hline(size_y[0] + 1, size_x[0], curses.ACS_HLINE, l_label - size_x[0] + 1)
            # Write message
            middle_x = (l_label - size_x[0] - len(self.message)) // 2
            middle_y = (size_y[1] - size_y[0]) // 2
            stdscr.addstr(size_y[0] + middle_y, size_x[0] + middle_x, self.message, curses.A_BOLD)

    def _plot_y_axis(self, stdscr, size_x, size_y, label=True):
        # Plot chart shape and labels
        displayY = size_y[1] - size_y[0] - 1
        label_x = size_x[1] - 5 if label else size_x[1]
        for point in range(displayY):
            if displayY != point:
                value_n = self.max_val / float(displayY) * float(displayY - point)
                try:
                    stdscr.addch(1 + size_y[0] + point, label_x, curses.ACS_LTEE)
                    if not label:
                        continue
                    if self.type_value == float:
                        lab_c = "{value:2.1f}{unit}".format(value=value_n, unit=self.unit)
                    else:
                        lab_c = "{value:3d}{unit}".format(value=int(value_n), unit=self.unit)
                    stdscr.addstr(1 + size_y[0] + point, label_x + 2, lab_c, curses.A_BOLD)
                except curses.error:
                    pass

    def _plot_x_axis(self, stdscr, size_x, size_y, label=True):
        displayX = size_x[1] - size_x[0] + 1
        val = float(displayX - 2) / float(len(self.values))
        # Draw ticks and labels
        ten_sec = int(self.tik * 1000 / self.refresh)
        counter = 0
        label_y = size_x[1] - 5 if label else size_x[1]
        # Draw line
        stdscr.hline(size_y[1] - 1, size_x[0], curses.A_UNDERLINE, label_y - size_x[0])
        for point in range(displayX):
            x_val = label_y - point
            if x_val >= size_x[0]:
                try:
                    # Draw tick label
                    if ((point) / ceil(val)) % ten_sec == 0:
                        stdscr.addch(size_y[1], x_val, curses.ACS_LLCORNER)
                    # Draw label
                    if counter > 0 and ((point - 1) / ceil(val)) % ten_sec == 0:
                        stdscr.addstr(size_y[1], x_val + 3, "-{time}s".format(time=self.tik * counter))
                    elif counter == 0 and ((point - 1) / ceil(val)) % ten_sec == 0:
                        if label:
                            stdscr.addstr(size_y[1], x_val + 3, "time")
                        stdscr.addstr(size_y[1], x_val + 1, "0")
                    # Increase counter
                    if ((point - 1) / ceil(val)) % ten_sec == 0:
                        counter += 1
                except curses.error:
                    pass

    def _plot_values(self, stdscr, size_x, size_y, label=True):
        # Area Plot data
        size_plot_x = [size_x[0], size_x[1] - 6 if label else size_x[1] - 1]
        size_plot_y = [size_y[0], size_y[1] - 1]
        size_y = size_plot_y[1] - size_plot_y[0]
        list_values = list(self.values)
        val = ceil(float(size_plot_x[1] - size_plot_x[0]) / float(len(list_values)))
        points = []
        for n in list_values:
            points += [n] * int(val)
        # Draw all chart
        color_step = len(self.color_chart)
        for idx, values in enumerate(reversed(points)):
            for chart_idx, value in enumerate(values):
                color_base = Chart.OFFSET_COLOR_CHART + self._color_obj_counter + chart_idx * color_step
                color_next = 1 if chart_idx != 0 else 0
                cell_val = value * size_y / self.max_val
                cell_val_int = int(cell_val)
                cell_val_mant = cell_val - cell_val_int
                if cell_val > 0 and size_plot_x[1] - idx >= size_plot_x[0]:
                    # Fill chart if request
                    # https://www.htmlsymbols.xyz/box-drawing
                    # Full block: \u2588 - 3/4 block \u2586 - Lower block: \u2584 - Small lower block: \u2581
                    if self.fill:
                        for n in range(cell_val_int - 1):
                            stdscr.addstr(size_plot_y[1] - n, size_plot_x[1] - idx, u'\u2588'.encode('utf-8'),
                                          curses.color_pair(color_base))
                        # Add head chart
                        if cell_val < 1.0:
                            stdscr.addstr(size_plot_y[1] - cell_val_int, size_plot_x[1] - idx, u'\u2581'.encode('utf-8'),
                                          curses.color_pair(color_base + color_next))
                        elif cell_val_mant == 0.0:
                            stdscr.addstr(size_plot_y[1] - cell_val_int + 1, size_plot_x[1] - idx, u'\u2584'.encode('utf-8'),
                                          curses.color_pair(color_base))
                        elif cell_val_mant <= 0.5:
                            stdscr.addstr(size_plot_y[1] - cell_val_int + 1, size_plot_x[1] - idx, u'\u2586'.encode('utf-8'),
                                          curses.color_pair(color_base + color_next))
                        elif cell_val_mant < 1.0:
                            stdscr.addstr(size_plot_y[1] - cell_val_int, size_plot_x[1] - idx, u'\u2581'.encode('utf-8'),
                                          curses.color_pair(color_base + color_next))
                            stdscr.addstr(size_plot_y[1] - cell_val_int + 1, size_plot_x[1] - idx, u'\u2588'.encode('utf-8'),
                                          curses.color_pair(color_base))
                    else:
                        stdscr.addstr(size_plot_y[1] - cell_val_int, size_plot_x[1] - idx, self.line, curses.color_pair(Chart.OFFSET_COLOR_TEXT))
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import os
import sys
# control command line
import curses
# Functions and decorators
from .colors import NColors
from functools import wraps


def set_xterm_title(title):
    '''
    Set XTerm title using escape sequences.
    By default, sets as 'Python' and the version number.
    '''
    # Make sure this terminal supports the OSC code (\33]),
    # though not necessarily that it supports setting the title.
    # If this check causes compatibility issues, you can add
    # items to the tuple, or remove the check entirely.
    if os.environ.get('TERM') in ('xterm',
                                  'xterm-color',
                                  'xterm-256color',
                                  'linux',
                                  'screen',
                                  'screen-256color',
                                  'screen-bce',
                                  ):
        sys.stdout.write('\33]0;' + title + '\a')
        sys.stdout.flush()


def jetson_clocks_gui(stdscr, offset, start, jetson):
    # Write status jetson_clocks
    jc_status_name = jetson.jetson_clocks.get_status()
    # Read status jetson_clocks
    if jc_status_name == "running":
        color = (curses.A_BOLD | NColors.green())  # Running (Bold)
    elif jc_status_name == "inactive":
        color = curses.A_NORMAL       # Normal (Grey)
    elif "ing" in jc_status_name:
        color = NColors.yellow()  # Warning (Yellow)
    else:
        color = NColors.red()  # Error (Red)
    # Show if JetsonClock is enabled or not
    if jetson.jetson_clocks.get_boot():
        jc_status_name = "[" + jc_status_name + "]"
    # Show status jetson_clocks
    plot_name_info(stdscr, offset, start, "Jetson Clocks", jc_status_name, color)


def check_curses(func):
    """ Check curses write """
    @wraps(func)
    def wrapped(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except curses.error:
            pass
    return wrapped


def strfdelta(tdelta, fmt):
    """ Print delta time
        - https://stackoverflow.com/questions/8906926/formatting-python-timedelta-objects
    """
    d = {"days": tdelta.days}
    d["hours"], rem = divmod(tdelta.seconds, 3600)
    d["minutes"], d["seconds"] = divmod(rem, 60)
    return fmt.format(**d)


def label_freq(frq, start='k'):
    szw, _, k_unit = size_min(frq, start=start)
    if szw >= 100:
        label = '{tot:2.0f}{unit}Hz'.format(tot=szw, unit=k_unit)
    elif szw >= 10:
        label = '{tot:2.0f} {unit}Hz'.format(tot=szw, unit=k_unit)
    else:
        label = '{tot:2.1f}{unit}Hz'.format(tot=szw, unit=k_unit)
    return label


def size_min(num, divider=1.0, n=0, start=''):
    return find_unit(num, 1024.0, divider, n, start)


def unit_min(num, divider=1.0, n=0, start=''):
    return find_unit(num, 1000.0, divider, n, start)


def find_unit(size, power, divider=1.0, n=0, start=''):
    n = 0
    power_labels = ['m', '', 'k', 'M', 'G', 'T']
    while size > power - 1:
        divider *= power
        size /= power
        n += 1
    idx = power_labels.index(start)
    return round(size, 1), divider, power_labels[n + idx]


def size_to_string(value, unit):
    return value_to_string(value, unit, "", size_min)


def unit_to_string(value, unit, type):
    return value_to_string(value, unit, type, unit_min)


def value_to_string(value, unit, type, func):
    value, _, unit = func(value, start=unit)
    value_string = str(value)
    if value >= 100:
        # value_string = value_string.rstrip('0').rstrip('.')
        value_string = value_string[:3].rstrip('.')
    return "{value}{unit}{type}".format(value=value_string, unit=unit, type=type)


def plot_dictionary(stdscr, pos_y, pos_x, name, data, size=None):
    size_y = 1
    size_x = 0
    try:
        stdscr.addstr(pos_y, pos_x, name, curses.A_BOLD)
    except curses.error:
        pass
    # Build table from dictionary
    for idx, (name, value) in enumerate(data.items()):
        # Plot nanme
        try:
            stdscr.addstr(pos_y + idx + 1, pos_x + 1, str(name) + ":", curses.A_BOLD)
        except curses.error:
            pass
        # Plot value
        color = curses.A_NORMAL if value else NColors.red()
        if not value:
            value = "MISSING"
        len_value = len(value)
        if size:
            if len(name) + len(value) + 3 > size:
                len_value = size - len(name) - 3
        try:
            stdscr.addstr(pos_y + idx + 1, pos_x + 3 + len(name), value[:len_value], color)
        except curses.error:
            pass
        size_x = max(size_x, len(name) + len(value) + 3)
        size_y += 1
    return size_y, size_x


def plot_name_info(stdscr, offset, start, name, value, color=curses.A_NORMAL, spacing=0):
    try:
        stdscr.addstr(offset, start, name + ":", curses.A_BOLD)
        stdscr.addstr(offset, start + len(name) + 2 + spacing, value, color)
    except curses.error:
        pass
    return len(name) + len(value) + 2
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import curses
# Page class definition
from .jtopgui import Page
from .lib.colors import NColors
from .lib.common import unit_to_string, plot_name_info
from .lib.linear_gauge import freq_gauge


def get_value_engine(engine):
    return unit_to_string(engine['cur'], 'k', 'Hz') if engine['online'] else '[OFF]'


def add_engine_in_list(label, engine, group, name):
    return [(label, get_value_engine(engine[group][name]))] if group in engine else []


def pass_orin(engine):
    return [
        add_engine_in_list('APE', engine, 'APE', 'APE') + add_engine_in_list('PVA0a', engine, 'PVA0', 'PVA0_CPU_AXI'),
        add_engine_in_list('DLA0c', engine, 'DLA0', 'DLA0_CORE') + add_engine_in_list('DLA1c', engine, 'DLA1', 'DLA1_CORE'),
        add_engine_in_list('NVENC', engine, 'NVENC', 'NVENC') + add_engine_in_list('NVDEC', engine, 'NVDEC', 'NVDEC'),
        add_engine_in_list('NVJPG', engine, 'NVJPG', 'NVJPG') + add_engine_in_list('NVJPG1', engine, 'NVJPG', 'NVJPG1'),
        add_engine_in_list('SE', engine, 'SE', 'SE') + add_engine_in_list('VIC', engine, 'VIC', 'VIC'),
    ]


def pass_orin_nano(engine):
    return [
        add_engine_in_list('APE', engine, 'APE', 'APE'),
        add_engine_in_list('NVENC', engine, 'NVENC', 'NVENC') + add_engine_in_list('NVDEC', engine, 'NVDEC', 'NVDEC'),
        add_engine_in_list('NVJPG', engine, 'NVJPG', 'NVJPG') + add_engine_in_list('NVJPG1', engine, 'NVJPG', 'NVJPG1'),
        add_engine_in_list('SE', engine, 'SE', 'SE') + add_engine_in_list('VIC', engine, 'VIC', 'VIC'),
    ]


def map_xavier(engine):
    return [
        add_engine_in_list('APE', engine, 'APE', 'APE') + add_engine_in_list('CVNAS', engine, 'CVNAS', 'CVNAS'),
        add_engine_in_list('DLA0c', engine, 'DLA0', 'DLA0_CORE') + add_engine_in_list('DLA1c', engine, 'DLA1', 'DLA1_CORE'),
        add_engine_in_list('NVENC', engine, 'NVENC', 'NVENC') + add_engine_in_list('NVDEC', engine, 'NVDEC', 'NVDEC'),
        add_engine_in_list('NVJPG', engine, 'NVJPG', 'NVJPG') + add_engine_in_list('PVA0a', engine, 'PVA0', 'PVA0_AXI'),
        add_engine_in_list('SE', engine, 'SE', 'SE') + add_engine_in_list('VIC', engine, 'VIC', 'VIC'),
    ]


def map_jetson_nano(engine):
    return [
        add_engine_in_list('APE', engine, 'APE', 'APE'),
        add_engine_in_list('NVENC', engine, 'NVENC', 'NVENC') + add_engine_in_list('NVDEC', engine, 'NVDEC', 'NVDEC'),
        add_engine_in_list('NVJPG', engine, 'NVJPG', 'NVJPG') + add_engine_in_list('SE', engine, 'SE', 'SE'),
    ]


MAP_JETSON_MODELS = {
    'orin nano': pass_orin_nano,
    'orin nx': pass_orin,
    'agx orin': pass_orin,
    'xavier': map_xavier,
    'jetson nano': map_jetson_nano,
    'nintendo': map_jetson_nano,
    'jetson tx': map_jetson_nano,
}


def engine_model(model):
    for name, func in MAP_JETSON_MODELS.items():
        if name.lower() in model.lower():
            return func
    return None


def map_engines(jetson):
    # Check if there is a map for each engine
    func_list_engines = engine_model(jetson.board['hardware']["Module"])
    try:
        if func_list_engines:
            return func_list_engines(jetson.engine)
    except KeyError:
        pass
    # Otherwise if not mapped show all engines
    list_engines = []
    for group in jetson.engine:
        list_engines += [[(name, get_value_engine(engine)) for name, engine in jetson.engine[group].items()]]
    return list_engines


def compact_engines(stdscr, pos_y, pos_x, width, height, jetson):
    center_x = pos_x + width // 2
    map_eng = map_engines(jetson)
    size_map = len(map_eng)
    # Write first line
    if size_map > 0:
        stdscr.addstr(pos_y, center_x - 7, " [HW engines] ", curses.A_BOLD)
        size_map += 1
    # Plot all engines
    size_table = 26
    for gidx, row in enumerate(map_eng):
        size_eng = size_table // len(row) - 1
        for idx, (name, value) in enumerate(row):
            if name is not None:
                color = curses.A_NORMAL if '[OFF]' in value else NColors.green() | curses.A_BOLD
                plot_name_info(stdscr, pos_y + gidx + 1, center_x - size_table // 2 + (size_eng + 1) * idx + 1, name, value, color=color)
    return size_map


class ENGINE(Page):

    def __init__(self, stdscr, jetson):
        super(ENGINE, self).__init__("ENG", stdscr, jetson)

    def draw(self, key, mouse):
        # Screen size
        height, width, first = self.size_page()
        # Draw all engines
        offset_y = first + 1
        offset_x = 1
        size_gauge = width - 2
        # Draw all engines
        for gidx, group in enumerate(self.jetson.engine):
            engines = self.jetson.engine[group]
            size_eng = size_gauge // len(engines) - 1
            # Plot all engines
            for idx, (name, engine) in enumerate(engines.items()):
                name_array = name.split("_")
                # Plot block name
                try:
                    if len(engines) > 1 and len(name_array) > 1:
                        self.stdscr.addstr(offset_y + gidx * 2, offset_x + (size_eng + 1) * idx,
                                           "{group}".format(group=group), NColors.cyan() | curses.A_BOLD)
                except curses.error:
                    pass
                # Plot Gauge
                new_name = ' '.join(name_array[1:]) if len(name_array) > 1 and len(engines) > 1 else name
                # Add name in plot string
                engine['name'] = new_name
                try:
                    freq_gauge(self.stdscr, offset_y + gidx * 2 + 1, offset_x + (size_eng + 1) * idx, size_eng, engine)
                except curses.error:
                    pass
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.


class JtopException(Exception):
    """
    raise when jtop fail. The message attached show the reason.
    """

    def __init__(self, message, errors=""):
        super(JtopException, self).__init__(message, errors)
        # Now for your custom code...
        self.message = message
        self.errors = errors

    def __repr__(self):
        return str(self.message)

    def __str__(self):
        return str(self.message)
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import curses
from .colors import NColors
from .common import unit_to_string


def basic_gauge(stdscr, pos_y, pos_x, size_w, data, bar='|'):
    """_summary_

        data = {
            'name': name value
            'color': color test
            'online': bool status
            'message': Offline message - Default OFF
            'coffline': Color offline - Default RED
            'values': [(value, color), (value, color), ... ] sum of values = 100
            'mleft': message on left
            'mright': message on right - Default percentage
        }
    """
    # Evaluate size without short name
    name_size = 0
    if 'name' in data:
        if data['name']:
            name = data['name']
            name_size = len(name) + 1
            # Show short name linear gauge
            name_color = data['color'] if 'color' in data else curses.A_NORMAL
            stdscr.addstr(pos_y, pos_x, name, name_color)
    # Size bar
    size_bar = size_w - name_size - 1
    # Draw gauge
    online = data['online'] if 'online' in data else True
    # Draw gauge border
    color_offline = data['coffline'] if 'coffline' in data else NColors.ired()
    stdscr.addstr(pos_y, pos_x + name_size, "[" + " " * size_bar + "]", curses.A_BOLD if online else color_offline)
    # Draw bar
    if online:
        values = data['values'] if 'values' in data else []
        # size dynamic bar
        total = min(sum([value for value, _ in values]), 100)
        n_bar = int(total) * size_bar // 100
        # Draw progress bar
        str_progress_bar = bar * n_bar + " " * (size_bar - n_bar)
        # Add label right otherwise write a percent
        label_right = data['mright'] if 'mright' in data else "{:.1f}%".format(total)
        if size_bar > len(label_right):
            str_progress_bar = str_progress_bar[:size_bar - len(label_right)] + label_right
        # Add message on left
        if 'mleft' in data:
            if size_bar > len(data['mleft']) + len(label_right):
                str_progress_bar = data['mleft'] + str_progress_bar[len(data['mleft']):]
        # Draw all values
        x_bar_start = 0
        old_val = 0
        for value, color in values:
            x_bar_end = int((old_val + value) * size_bar) // 100
            stdscr.addstr(pos_y, pos_x + name_size + x_bar_start + 1, str_progress_bar[x_bar_start:x_bar_end], color)
            x_bar_start = x_bar_end
            old_val += int(value)
        # Draw grey part or message
        grey_part = str_progress_bar[x_bar_start:]
        stdscr.addstr(pos_y, pos_x + name_size + x_bar_start + 1, grey_part, curses.A_DIM)
    else:
        # Show message status
        try:
            stdscr.addstr(pos_y, pos_x + name_size + 2, data['message'] if 'message' in data else "OFF", color_offline)
        except curses.error:
            pass


def basic_gauge_simple(stdscr, pos_y, pos_x, size, freq_data, unit='k'):
    # Name gauge
    name = freq_data['name'] if 'name' in freq_data else ""
    # Current value in string
    curr_string = unit_to_string(freq_data['cur'], unit, 'Hz')
    # Draw name engine
    stdscr.addstr(pos_y, pos_x, name, NColors.cyan())
    # Write online bar
    size_bar = size - len(name) - len(curr_string) - 4
    start_bar = pos_x + len(name) + 1 if len(name) > 0 else pos_x
    end_bar = start_bar + size_bar
    # Check if there is a limit
    color_bar = NColors.green() if freq_data['online'] else NColors.red()
    if freq_data['online']:
        stdscr.hline(pos_y, start_bar + 1, curses.ACS_HLINE, size_bar)
        stdscr.addch(pos_y, start_bar + size_bar, curses.ACS_DIAMOND, curses.A_BOLD)
        if size_bar > 7:
            stdscr.addstr(pos_y, end_bar - (size) // 2, " RUNNING ", color_bar | curses.A_BOLD)
    else:
        stdscr.hline(pos_y, start_bar + 1, curses.ACS_BULLET, size_bar)
        if size_bar > 7:
            stdscr.addstr(pos_y, start_bar + (size_bar - 5) // 2, ' OFF ', color_bar | curses.A_NORMAL)
        else:
            stdscr.addstr(pos_y, start_bar + (size_bar - 3) // 2, 'OFF', color_bar | curses.A_NORMAL)
    # Draw current frequency
    stdscr.addstr(pos_y, pos_x + size - 6, curr_string, NColors.italic())


def freq_gauge(stdscr, pos_y, pos_x, size, freq_data):
    # Name gauge
    name = freq_data['name'] if 'name' in freq_data else ""
    # Current value in string
    curr_string = unit_to_string(freq_data['cur'], 'k', 'Hz')
    # If there is a min and a max
    if 'max' in freq_data:
        value = ((freq_data['cur'] - freq_data['min']) / (freq_data['max'] - freq_data['min'])) * 100 if freq_data['min'] != freq_data['max'] else 100
        # Convert values data
        data = {
            'name': name,
            'color': NColors.cyan(),
            'online': freq_data['online'] if 'online' in freq_data else True,
            'values': [(value, NColors.green())],
            'mleft': unit_to_string(freq_data['min'], 'k', 'Hz') if 'min' in freq_data else "",
            'mright': unit_to_string(freq_data['max'], 'k', 'Hz') if 'max' in freq_data else "",
        }
        basic_gauge(stdscr, pos_y, pos_x, size - 8, data, bar=":")
        # Draw current frequency
        stdscr.addstr(pos_y, pos_x + size - 6, curr_string, NColors.italic())
    else:
        basic_gauge_simple(stdscr, pos_y, pos_x, size, freq_data)
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import json
import sys
import copy
# Create logger
logger = logging.getLogger(__name__)
JTOP_DATA_FOLDER = 'local/jtop'


def make_config_service(data_folder=JTOP_DATA_FOLDER):
    path = get_config_service(data_folder)
    if not os.path.isdir(path):
        logger.info("Build service folder in {path}".format(path=path))
        # Make folder directory
        os.makedirs(path)


def get_config_service(data_folder=JTOP_DATA_FOLDER):
    path = sys.prefix
    if hasattr(sys, 'real_prefix'):
        path = sys.real_prefix
    if hasattr(sys, 'base_prefix'):
        path = sys.base_prefix
    # Return directory folder
    return "{path}/{data_folder}".format(path=path, data_folder=data_folder)


class Config:

    def __init__(self):
        # Build folder if doesn't exists
        make_config_service()
        # Load configuration path
        self.config_file = self.path + '/config.json'
        # Load configuration
        self._config = self._load()
        self._last_config = copy.deepcopy(self._config)

    def set(self, instance, default=None):
        # Update configuration
        self._config[instance] = default
        # Store configuration
        if self._last_config != self._config:
            self._store()
            # Update last configuration
            self._last_config = copy.deepcopy(self._config)

    def get(self, instance, default=None):
        return self._config.get(instance, default)

    @property
    def path(self):
        return get_config_service()

    def _load(self):
        config = {}
        # Load configuration if exist
        if not os.path.isfile(self.config_file):
            return config
        logger.info("Load config from {path}".format(path=self.config_file))
        with open(self.config_file) as json_file:
            config = json.load(json_file)
        return config

    def _store(self):
        logger.info("Store config to {path}".format(path=self.config_file))
        # Write configuration
        with open(self.config_file, 'w') as outfile:
            json.dump(self._config, outfile, sort_keys=True, indent=4)

    def clear(self):
        self._config = {}
        self._last_config = {}
        if os.path.isfile(self.config_file):
            logger.info("Clear config in {path}".format(path=self.config_file))
            # Remove configuration file
            os.remove(self.config_file)
            return True
        return False

    def items(self):
        return self._config.items()

    def keys(self):
        return self._config.keys()

    def values(self):
        return self._config.values()

    def __contains__(self, key):
        return key in self._config

    def __repr__(self):
        return repr(self._config)

    def __str__(self):
        return str(self._config)
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import curses
from .jtopgui import Page
# Graphics elements
from .lib.common import (
    strfdelta,
    plot_name_info,
    jetson_clocks_gui)
from .lib.colors import NColors
from .lib.linear_gauge import basic_gauge
from .lib.common import size_to_string
from .lib.process_table import ProcessTable
from .pcpu import compact_cpus
from .pgpu import compact_gpu
from .pmem import compact_memory
from .pengine import compact_engines
from .pcontrol import compact_temperatures, compact_power


def compact_status(stdscr, pos_y, pos_x, width, height, jetson):
    line_counter = 0
    # Fan status
    if jetson.fan:
        for name, fan in jetson.fan.items():
            for idx, speed in enumerate(fan['speed']):
                try:
                    data = {
                        'name': 'Fan {idx}'.format(idx=idx) if len(fan['speed']) > 1 else 'FAN',
                        'color': NColors.cyan(),
                        'online': jetson.fan,
                        'values': [(speed, NColors.cyan())]
                    }
                    # Show RPM if exist
                    if 'rpm' in fan:
                        rpm = "{rpm:>4}RPM".format(rpm=fan['rpm'][idx])
                        stdscr.addstr(pos_y + line_counter, pos_x + width - 9, rpm, curses.A_NORMAL)
                    size_fan_gauge = width - 12 if 'rpm' in fan else width - 3
                    basic_gauge(stdscr, pos_y + line_counter, pos_x + 1, size_fan_gauge, data)
                    line_counter += 1
                except curses.error:
                    pass
    else:
        try:
            data = {
                'name': 'Fan',
                'color': NColors.cyan(),
                'online': False,
                'coffline': NColors.icyan(),
                'message': 'NOT AVAILABLE',
            }
            basic_gauge(stdscr, pos_y + line_counter, pos_x + 1, width - 3, data)
            line_counter += 1
        except curses.error:
            pass
    # Jetson clocks status: Running (Green) or Normal (Grey)
    if jetson.jetson_clocks is not None:
        jetson_clocks_gui(stdscr, pos_y + line_counter, pos_x + 1, jetson)
        line_counter += 1
    # NVP Model
    if jetson.nvpmodel is not None:
        plot_name_info(stdscr, pos_y + line_counter, pos_x + 1, "NV Power[" + str(jetson.nvpmodel.id) + "]", jetson.nvpmodel.name)
        line_counter += 1
    # Model board information
    uptime_string = strfdelta(jetson.uptime, "{days} days {hours}:{minutes}:{seconds}")
    plot_name_info(stdscr, pos_y + line_counter, pos_x + 1, "Uptime", uptime_string)
    return line_counter + 1


def disk_gauge(stdscr, pos_y, pos_x, size, disk_status):
    # value disk
    value = int(float(disk_status['used']) / float(disk_status['total']) * 100.0)
    used = size_to_string(disk_status['used'], disk_status['unit'])
    total = size_to_string(disk_status['total'], disk_status['unit'])
    data = {
        'name': 'Dsk',
        'color': NColors.yellow(),
        'values': [(value, NColors.yellow())],
        'mright': "{used}/{total}".format(used=used, total=total)
    }
    basic_gauge(stdscr, pos_y, pos_x, size - 2, data, bar="#")
    return 1


class ALL(Page):

    def __init__(self, stdscr, jetson):
        super(ALL, self).__init__("ALL", stdscr, jetson)
        # Add Process table
        if self.jetson.processes:
            self.process_table = ProcessTable(self.stdscr, self.jetson)
        # Number columns
        self._max_height_menu = 0
        self._columns = []
        if jetson.engine:
            self._columns += [compact_engines]
            self._max_height_menu = max(self._max_height_menu, 5)
        if jetson.temperature:
            self._columns += [compact_temperatures]
            self._max_height_menu = max(self._max_height_menu, len(jetson.temperature))
        if jetson.power:
            self._columns += [compact_power]
            self._max_height_menu = max(self._max_height_menu, len(jetson.power['rail']) + 1)
        # End corner
        self._max_height_menu += 2

    def draw(self, key, mouse):
        """
            Update screen with values
        """
        # Screen size
        height, width, first = self.size_page()
        line_counter = first + 1
        # Plot Status CPU
        line_counter += compact_cpus(self.stdscr, line_counter, 0, width, self.jetson)
        # Plot status memory
        size_memory = compact_memory(self.stdscr, line_counter, 0, width // 2, height, self.jetson)
        # Plot compact info
        size_status = compact_status(self.stdscr, line_counter, width // 2, width // 2, height, self.jetson)
        # Update line counter
        line_counter += max(size_memory, size_status)
        # GPU linear gauge info
        if height > line_counter:
            line_counter += compact_gpu(self.stdscr, line_counter, 0, width, self.jetson)
        # Status disk
        if height > line_counter:
            line_counter += disk_gauge(self.stdscr, line_counter, 0, width, self.jetson.disk)
        # Plot all processes
        height_free_area = height - line_counter - self._max_height_menu - 1
        offset_process_y = 0
        if self.jetson.processes:
            line_counter += self.process_table.draw(line_counter, 0, width, height_free_area, key, mouse)
            offset_process_y = 2
        # Evaluate number of columns
        n_columns = len(self._columns)
        # If empty return
        if n_columns == 0:
            return
        # Plot low bar background line
        pos_y_mini_menu = line_counter + offset_process_y
        if height_free_area > offset_process_y:
            pos_y_mini_menu = height - self._max_height_menu - 1
        column_height = height - pos_y_mini_menu
        if column_height <= 1:
            return
        # Upper block
        try:
            self.stdscr.addch(pos_y_mini_menu, 0, curses.ACS_ULCORNER)
            self.stdscr.addch(pos_y_mini_menu, width - 1, curses.ACS_URCORNER)
            self.stdscr.hline(pos_y_mini_menu, 1, curses.ACS_HLINE, width - 2)
        except curses.error:
            pass
        # vertical Lines
        if column_height > 3:
            self.stdscr.vline(pos_y_mini_menu + 1, 0, curses.ACS_VLINE, column_height - 3)
            self.stdscr.vline(pos_y_mini_menu + 1, width - 1, curses.ACS_VLINE, column_height - 3)
        # Lower line
        try:
            self.stdscr.addch(pos_y_mini_menu + self._max_height_menu - 1, 0, curses.ACS_LLCORNER)
            self.stdscr.addch(pos_y_mini_menu + self._max_height_menu - 1, width - 1, curses.ACS_LRCORNER)
            self.stdscr.hline(pos_y_mini_menu + self._max_height_menu - 1, 1, curses.ACS_HLINE, width - 2)
        except curses.error:
            self.stdscr.addch(pos_y_mini_menu + column_height - 2, 0, curses.ACS_DARROW, curses.A_REVERSE | curses.A_BOLD)
            self.stdscr.addch(pos_y_mini_menu + column_height - 2, width - 1, curses.ACS_DARROW, curses.A_REVERSE | curses.A_BOLD)
        # Cancel columns depend of the width size
        n_print = n_columns
        if width < 75 and n_columns > 1:
            n_print -= 1
        if width < 49 and n_columns > 2:
            n_print -= 1
        # Add engines, temperature and power
        column_width = (width) // (n_print)
        for nline in range(n_print):
            func = self._columns[nline]
            func(self.stdscr, pos_y_mini_menu, column_width * nline, column_width, column_height, self.jetson)
            if nline < n_print - 1:
                self.stdscr.addch(pos_y_mini_menu, column_width * (nline + 1), curses.ACS_TTEE)
                if column_height > 3:
                    self.stdscr.vline(pos_y_mini_menu + 1, column_width * (nline + 1), curses.ACS_VLINE, column_height - 3)
                if height_free_area >= offset_process_y:
                    self.stdscr.addch(pos_y_mini_menu + column_height - 2, column_width * (nline + 1), curses.ACS_BTEE)
                else:
                    self.stdscr.addch(pos_y_mini_menu + column_height - 2, column_width * (nline + 1),
                                      curses.ACS_DARROW, curses.A_REVERSE | curses.A_BOLD)
# EOF



# -*- coding: UTF-8 -*-
# This file is part of the jetson_stats package (https://github.com/rbonghi/jetson_stats or http://rnext.it).
# Copyright (c) 2019-2023 Raffaello Bonghi.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import curses
from .jtopgui import Page
# Graphics elements
from .lib.common import unit_to_string
# Graphic library
from .lib.colors import NColors
from .lib.chart import Chart
from .lib.smallbutton import SmallButton, ButtonList

FAN_STEP = 10
PROFILE_STR = "Profiles:"
TEMPERATURE_MAX = 84
TEMPERATURE_CRIT = 100


def color_temperature(stdscr, pos_y, pos_x, name, sensor, offset=0):
    if not sensor['online']:
        stdscr.addstr(pos_y, pos_x, name)
        stdscr.addstr(pos_y, pos_x + offset + 5, "Offline", NColors.yellow())
        return
    # Print temperature name
    temperature = sensor['temp']
    # Set color temperature
    max_value = sensor['max'] if 'max' in sensor else TEMPERATURE_MAX
    crit_value = sensor['crit'] if 'crit' in sensor else TEMPERATURE_CRIT
    # Set color
    color = curses.A_NORMAL
    if temperature >= crit_value:
        color = NColors.red()
    elif temperature >= max_value:
        color = NColors.yellow()
    # Print temperature value
    stdscr.addstr(pos_y, pos_x, name)
    stdscr.addstr(pos_y, pos_x + offset + 5, ("{val:3.2f}C").format(val=temperature), color)


def compact_temperatures(stdscr, pos_y, pos_x, width, height, jetson):
    counter = 0
    center_x = pos_x + width // 2 + 1
    offset = 2
    # Plot title
    stdscr.addstr(pos_y, center_x - offset - 10, " [Sensor] ", curses.A_BOLD)
    stdscr.addstr(pos_y, center_x + offset, " [Temp] ", curses.A_BOLD)
    # Plot name and temperatures
    for idx, (name, sensor) in enumerate(jetson.temperature.items()):
        # Plot temperature
        try:
            color_temperature(stdscr, pos_y + idx + 1, center_x - 5 * offset, name, sensor, offset=4 * offset)
        except curses.error:
            break
        counter = idx
    return counter


def compact_power(stdscr, pos_y, pos_x, width, height, jetson):
    LIMIT = 25
    # center_x = pos_x + width // 2 if width > LIMIT else pos_x + width // 2 + 4
    center_x = pos_x + width // 2 + 2 if width > LIMIT else pos_x + width // 2 + 6
    column_power = 9
    # Plot title
    stdscr.addstr(pos_y, center_x - column_power - 5, " [Power] ", curses.A_BOLD)
    stdscr.addstr(pos_y, center_x - 3, " [Inst] ", curses.A_BOLD)
    if width > LIMIT:
        stdscr.addstr(pos_y, center_x + column_power - 4, " [Avg] ", curses.A_BOLD)
    # Plot watts
    power = jetson.power['rail']
    for idx, name in enumerate(power):
        if idx + 1 >= height:
            return idx
        value = power[name]
        string_name = name.replace("VDDQ_", "").replace("VDD_", "").replace("_", " ")
        stdscr.addstr(pos_y + idx + 1, center_x - column_power - 5, string_name, curses.A_NORMAL)
        unit_power = unit_to_string(value['power'], 'm', 'W')
        stdscr.addstr(pos_y + idx + 1, center_x - 1, unit_power, curses.A_NORMAL)
        if width > LIMIT:
            unit_avg = unit_to_string(value['avg'], 'm', 'W')
            stdscr.addstr(pos_y + idx + 1, center_x + column_power - 3, unit_avg, curses.A_NORMAL)
    # Plot totals before finishing
    total = jetson.power['tot']
    len_power = len(power)
    if len_power + 1 >= height:
        return len(power)
    name_total = total['name'] if 'name' in total else 'ALL'
    stdscr.addstr(pos_y + len_power + 1, center_x - column_power - 5, name_total, curses.A_BOLD)
    unit_power = unit_to_string(total['power'], 'm', 'W')
    stdscr.addstr(pos_y + len_power + 1, center_x - 1, unit_power, curses.A_BOLD)
    if width > LIMIT:
        unit_avg = unit_to_string(total['avg'], 'm', 'W')
        stdscr.addstr(pos_y + len_power + 1, center_x + column_power - 3, unit_avg, curses.A_BOLD)
    return len(power) + 1


class CTRL(Page):

    def __init__(self, stdscr, jetson):
        super(CTRL, self).__init__("CTRL", stdscr, jetson)
        # Only if exist a fan will be load a chart
        # Initialize FAN chart
        self._fan_gui = {}
        for fan_name in self.jetson.fan:
            fan = self.jetson.fan[fan_name]
            # Initialize profile and list of fan
            profiles = jetson.fan.all_profiles(fan_name)
            button_list = ButtonList(stdscr, self.action_fan_profile, profiles, info={'name': fan_name})
            size_profile = max([len(profile) for profile in profiles] + [len(PROFILE_STR)]) + 2
            self._fan_gui[fan_name] = {'profile': button_list, 'fan': [], 'size_w': size_profile, 'len_profiles': len(profiles)}
            # Initialize all fan chart and buttons
            for idx in range(len(fan['speed'])):
                chart_fan = Chart(jetson, "{name} {idx}".format(name=fan_name.upper(), idx=idx), self.update_chart,
                                  line="o", color_text=curses.COLOR_CYAN, color_chart=[curses.COLOR_CYAN])
                button_increase = SmallButton(stdscr, self.action_fan_increase, info={'name': fan_name, 'idx': idx})
                button_decrease = SmallButton(stdscr, self.action_fan_decrease, info={'name': fan_name, 'idx': idx})
                self._fan_gui[fan_name]['fan'] += [{'chart': chart_fan, 'increase': button_increase, 'decrease': button_decrease}]
        # Initialize jetson_clocks buttons
        if self.jetson.jetson_clocks is not None:
            self._jetson_clocks_start = SmallButton(stdscr, self.action_jetson_clocks_start, trigger_key='s')
            self._jetson_clocks_boot = SmallButton(stdscr, self.action_jetson_clocks_boot, trigger_key='e')
        # Initialize NVP Model buttons
        if self.jetson.nvpmodel is not None:
            self._nvp_default = self.jetson.nvpmodel.get_default()
            # nvp_modes = [name.replace('MODE_', '').replace('_', ' ') for name in self.jetson.nvpmodel.modes]
            self._nvpmodel_profile = ButtonList(stdscr, self.action_nvpmodels, self.jetson.nvpmodel.models)
            self._nvpmodel_increase = SmallButton(stdscr, self.action_nvp_increase, trigger_key='+')
            self._nvpmodel_decrease = SmallButton(stdscr, self.action_nvp_decrease, trigger_key='-')

    def action_fan_profile(self, info, selected):
        # Set new fan profile
        self.jetson.fan.set_profile(info['name'], info['label'])

    def action_fan_increase(self, info, selected):
        # Read current speed
        speed = self.jetson.fan.get_speed(info['name'], info['idx'])
        # Round and increase speed
        spd = round(speed / 10) * 10 + FAN_STEP
        new_speed = spd if spd <= 100 else 100
        # Update fan speed
        self.jetson.fan.set_speed(info['name'], new_speed, info['idx'])

    def action_fan_decrease(self, info, selected):
        # Read current speed
        speed = self.jetson.fan.get_speed(info['name'], info['idx'])
        # Round and decrease speed
        spd = round(speed / 10) * 10 - FAN_STEP
        new_speed = spd if spd >= 0 else 0
        # Update fan speed
        self.jetson.fan.set_speed(info['name'], new_speed, info['idx'])

    def action_jetson_clocks_start(self, info, selected):
        # Start jetson_clocks
        self.jetson.jetson_clocks = not self.jetson.jetson_clocks

    def action_jetson_clocks_boot(self, info, selected):
        # Start jetson_clocks
        self.jetson.jetson_clocks.boot = not self.jetson.jetson_clocks.boot

    def action_nvpmodels(self, info, selected):
        # Set new nvpmodel
        self.jetson.nvpmodel = info['label']

    def action_nvp_increase(self, info, selected):
        # NVPmodel controller
        if self.jetson.nvpmodel.id >= len(self.jetson.nvpmodel.models) - 1:
            return
        self.jetson.nvpmodel += 1

    def action_nvp_decrease(self, info, selected):
        # NVPmodel controller
        if self.jetson.nvpmodel.id <= 0:
            return
        self.jetson.nvpmodel -= 1

    def update_chart(self, jetson, name):
        info_chart = name.split(" ")
        name = info_chart[0].lower()
        idx = int(info_chart[1])
        speed = jetson.fan[name]['speed'][idx]
        # Append in list
        return {
            'value': [speed],
        }

    def control_jetson_clocks(self, pos_y, pos_x, key, mouse):
        # Show jetson_clocks
        try:
            self.stdscr.addstr(pos_y, pos_x, "Jetson Clocks:", curses.A_BOLD)
        except curses.error:
            pass
        # Draw jetson_clocks status button
        try:
            # Status jetson_clocks
            jetson_clocks_status = self.jetson.jetson_clocks.status
            # Color status
            if jetson_clocks_status == "running":
                color = (curses.A_BOLD | NColors.green())  # Running (Bold)
            elif jetson_clocks_status == "inactive":
                color = curses.A_NORMAL       # Normal (Grey)
            elif "ing" in jetson_clocks_status:
                color = NColors.yellow()  # Warning (Yellow)
            else:
                color = NColors.red()  # Error (Red)
            self._jetson_clocks_start.update(pos_y, pos_x + 15, jetson_clocks_status, key, mouse, color=color)
        except curses.error:
            pass
        # Draw boot button
        try:
            self.stdscr.addstr(pos_y, pos_x + 32, "on boot:", curses.A_BOLD)
        except curses.error:
            pass
        try:
            boot = self.jetson.jetson_clocks.boot
            jetson_clocks_boot = "enable" if boot else "disable"
            color_boot = NColors.green() if boot else curses.A_NORMAL
            self._jetson_clocks_boot.update(pos_y, pos_x + 40, jetson_clocks_boot, key, mouse, color=color_boot)
        except curses.error:
            pass

    def control_nvpmodes(self, pos_y, pos_x, key, mouse):
        # Draw all profiles
        try:
            self.stdscr.addstr(pos_y, pos_x, "NVP modes:", curses.A_BOLD)
        except curses.error:
            pass
        # Write ID NVP model
        id = self.jetson.nvpmodel.id
        color = NColors.yellow() if self.jetson.nvpmodel.is_running() else curses.A_BOLD
        try:
            self.stdscr.addstr(pos_y, pos_x + 16, str(id), color)
        except curses.error:
            pass
        # Add buttons -/+
        try:
            self._nvpmodel_decrease.update(pos_y, pos_x + 11, key=key, mouse=mouse)
            self._nvpmodel_increase.update(pos_y, pos_x + 18, key=key, mouse=mouse)
        except curses.error:
            pass
        # Draw all modes
        current_mode = self.jetson.nvpmodel.name
        colors = [curses.A_NORMAL if status else NColors.red() for status in self.jetson.nvpmodel.status]
        try:
            self._nvpmodel_profile.update(pos_y + 1, pos_x + 2, key, mouse, current_mode, colors)
        except curses.error:
            pass
        try:
            # Write letter D for default
            self.stdscr.addstr(pos_y + self._nvp_default['id'] + 1, pos_x, "D", curses.A_BOLD)
        except curses.error:
            pass

    def control_power(self, pos_y, pos_x, key, mouse):
        if not self.jetson.power:
            return
        # Width  table
        width = 53
        # Draw all power
        power = self.jetson.power['rail']
        # Draw head table
        try:
            self.stdscr.addch(pos_y, pos_x, curses.ACS_ULCORNER)
            self.stdscr.addch(pos_y, pos_x + width - 1, curses.ACS_URCORNER)
            self.stdscr.hline(pos_y, pos_x + 1, curses.ACS_HLINE, width - 2)
            self.stdscr.addstr(pos_y, pos_x + 5, " Power ", curses.A_BOLD)
        except curses.error:
            pass
        # Draw header table
        try:
            self.stdscr.addstr(pos_y + 1, pos_x, "[Name]", curses.A_BOLD)
            self.stdscr.addstr(pos_y + 1, pos_x + 18, "[Power]", curses.A_BOLD)
            self.stdscr.addstr(pos_y + 1, pos_x + 26, "[Volt]", curses.A_BOLD)
            self.stdscr.addstr(pos_y + 1, pos_x + 33, "[Curr]", curses.A_BOLD)
            self.stdscr.addstr(pos_y + 1, pos_x + 40, "[Warn]", curses.A_BOLD)
            self.stdscr.addstr(pos_y + 1, pos_x + 47, "[Crit]", curses.A_BOLD)
        except curses.error:
            pass
        # Draw all values
        pos_y_table = pos_y + 2
        for idx, name in enumerate(power):
            value = power[name]
            try:
                self.stdscr.addstr(pos_y_table + idx, pos_x, name, curses.A_NORMAL)
            except curses.error:
                pass
            # Convert all values in readable strings
            unit_volt = unit_to_string(value['volt'], 'm', 'V')
            unit_curr = unit_to_string(value['curr'], 'm', 'A')
            unit_power = unit_to_string(value['power'], 'm', 'W')
            # Print all values
            try:
                self.stdscr.addstr(pos_y_table + idx, pos_x + 18, unit_power, curses.A_NORMAL)
                self.stdscr.addstr(pos_y_table + idx, pos_x + 26, unit_volt, curses.A_NORMAL)
                self.stdscr.addstr(pos_y_table + idx, pos_x + 33, unit_curr, curses.A_NORMAL)
            except curses.error:
                pass
            if 'warn' in value:
                try:
                    unit_curr_warn = unit_to_string(value['warn'], 'm', 'A')
                    self.stdscr.addstr(pos_y_table + idx, pos_x + 40, unit_curr_warn, curses.A_NORMAL)
                except curses.error:
                    pass
            if 'crit' in value:
                try:
                    unit_curr_crit = unit_to_string(value['crit'], 'm', 'A')
                    self.stdscr.addstr(pos_y_table + idx, pos_x + 47, unit_curr_crit, curses.A_NORMAL)
                except curses.error:
                    pass
        # Draw total power
        total = self.jetson.power['tot']
        len_power = len(power)
        try:
            name_total = total['name'] if 'name' in total else 'ALL'
            self.stdscr.addstr(pos_y_table + len_power, pos_x, name_total, curses.A_BOLD)
        except curses.error:
            pass
        try:
            unit_power_total = unit_to_string(total['power'], 'm', 'W')
            self.stdscr.addstr(pos_y_table + len_power, pos_x + 18, unit_power_total, curses.A_BOLD)
        except curses.error:
            pass
        try:
            if 'volt' in total:
                unit_volt = unit_to_string(total['volt'], 'm', 'V')
                self.stdscr.addstr(pos_y_table + len_power, pos_x + 26, unit_volt, curses.A_BOLD)
        except curses.error:
            pass
        try:
            if 'curr' in total:
                unit_curr = unit_to_string(total['curr'], 'm', 'A')
                self.stdscr.addstr(pos_y_table + len_power, pos_x + 33, unit_curr, curses.A_BOLD)
        except curses.error:
            pass
        try:
            if 'warn' in total:
                unit_curr_warn = unit_to_string(total['warn'], 'm', 'A')
                self.stdscr.addstr(pos_y_table + len_power, pos_x + 40, unit_curr_warn, curses.A_BOLD)
        except curses.error:
            pass
        try:
            if 'crit' in total:
                unit_curr_crit = unit_to_string(total['crit'], 'm', 'A')
                self.stdscr.addstr(pos_y_table + len_power, pos_x + 47, unit_curr_crit, curses.A_BOLD)
        except curses.error:
            pass

    def draw(self, key, mouse):
        # Screen size
        height, width, first = self.size_page()
        # Measure height
        fan_height = (height * 1 // 3 + 2) // len(self.jetson.fan) if len(self.jetson.fan) > 0 else 0
        # Draw all GPU
        for fan_idx, (fan_gui, fan_name) in enumerate(zip(self._fan_gui, self.jetson.fan)):
            gui_chart = self._fan_gui[fan_gui]
            fan = self.jetson.fan[fan_name]
            num_fans = len(fan['speed'])
            # Print all profiles
            try:
                pos_y_profiles = fan_height // 2 - gui_chart['len_profiles']
                size_profile = gui_chart['size_w']
                self.stdscr.addstr(first + 1 + fan_idx * (fan_height + 1) + pos_y_profiles - 1, 1, PROFILE_STR, curses.A_BOLD)
            except curses.error:
                pass
            # Split width for each pwm
            fan_speed_width = (width - size_profile - 6) // num_fans
            # Draw a button list with all profiles
            try:
                profile = self.jetson.fan.get_profile(fan_name)
                gui_chart['profile'].update(first + 1 + fan_idx * (fan_height + 1) + pos_y_profiles, 1, key, mouse, profile)
            except curses.error:
                pass
            # Print all fans
            for idx, speed in enumerate(fan['speed']):
                # Set size chart gpu
                size_x = [size_profile + idx * fan_speed_width, size_profile + (idx + 1) * (fan_speed_width - 1)]
                size_y = [first + 1 + fan_idx * (fan_height + 1), first + 1 + (fan_idx + 1) * (fan_height - 1)]
                # Print speed and RPM
                label_fan = "PWM {speed: >3.0f}%".format(speed=speed)
                if 'rpm' in fan:
                    label_fan += " - {rpm}RPM".format(rpm=fan['rpm'][idx])
                # Draw GPU chart
                gui_chart['fan'][idx]['chart'].draw(self.stdscr, size_x, size_y, label=label_fan, y_label=False)
                # Draw speed buttons
                pos_x_control_fan = (fan_speed_width - 6) // 2
                if fan_speed_width > 40:
                    self.stdscr.addstr(first + 1 + fan_idx * (fan_height + 1),
                                       size_profile + idx * fan_speed_width + pos_x_control_fan + 4,
                                       "Speed", curses.A_BOLD)
                try:
                    gui_chart['fan'][idx]['decrease'].update(first + 1 + fan_idx * (fan_height + 1),
                                                             size_profile + idx * fan_speed_width + pos_x_control_fan + 10,
                                                             '-', key, mouse)
                    gui_chart['fan'][idx]['increase'].update(first + 1 + fan_idx * (fan_height + 1),
                                                             size_profile + idx * fan_speed_width + pos_x_control_fan + 14,
                                                             '+', key, mouse)
                except curses.error:
                    pass
            # Plot y axis
            gui_chart['fan'][0]['chart'].draw_y_axis(self.stdscr,
                                                     first + 1 + fan_idx * (fan_height + 1),
                                                     size_profile + num_fans * (fan_speed_width - 1) + 1,
                                                     fan_height - 1)
        # Draw jetson clocks
        line_counter = fan_height
        if self.jetson.jetson_clocks is not None:
            self.control_jetson_clocks(first + 1 + line_counter, 1, key, mouse)
            line_counter += 1
        # Draw nvpmodels
        width_spacing = 5
        if self.jetson.nvpmodel is not None:
            self.control_nvpmodes(first + 1 + line_counter, 1, key, mouse)
            width_spacing = width // 2 - 16
        # Draw all power info
        self.control_power(first + 1 + line_counter, width_spacing, key, mouse)
# EOF
